#! /usr/bin/env python

"""Script that calls the jwst_download.py module. This is the top-level script that users should call.
It will parse arguments, make tables, query MAST, download files, and create webpages.
"""
import sys
from jwst_mast_query import jwst_download as jdl


def main():
    download = jdl.download_mast()
    parser = download.define_options()
    args = parser.parse_args()

    # The config file is loaded into self.params, and overwritten with the arguments that are not None
    download.get_arguments(args)
    if download.verbose>2:
        print('params:')
        for key, value in download.params.items():
            print(key, value)

    # Use arguments or $API_MAST_TOKEN to login
    download.login(raiseErrorFlag=True)

    # self.outrootdir is set depending on outrootdir and outsubdir in cfg file or through the options --outrootdir and --outsubdir
    download.set_outrootdir()
    if download.verbose: print(f'Outdir: {download.outrootdir}')

    # Make the tables, but don't show them yet, since the output files need to be updated first
    if download.mk_all_tables(showtables=True):
        sys.exit(0)

    if not download.params['skipdownload'] and len(download.ix_selected_products)>0:
        download.download_products()
        print('\n######################\n### Downloaded Selected Products:\n######################')
        download.productTable.write(indices=download.ix_selected_products,columns=download.params['outcolumns_productTable'])
    else:
        if len(download.ix_selected_products)<=0:
            print('############## Nothing selected!!!! exiting...')
            sys.exit(0)

    # Make the webpages
    if download.params['makewebpages']:
        download.mk_webpages4propIDs()


if __name__ == '__main__':
    main()


#!/usr/bin/env python
'''
wrapper around pandas with convenience functions to ease handling of tables
A. Rest
'''
import sys,os,re,types,copy,io
import numpy as np
from astropy.time import Time
import astropy.io.fits as fits
import pandas as pd
from pandas.core.dtypes.common import is_object_dtype,is_float_dtype,is_string_dtype,is_integer_dtype

from astropy.nddata import bitmask
from astropy import units as u
from astropy.coordinates import SkyCoord

from scipy.interpolate import interp1d

def split_commonpath(path1,path2, strip_basenames=False, raiseError=True):
    """
    compare two paths, and return the common path, and whatever is left from
    path1 that is not common

    Parameters
    ----------
    path1 : string
    path2 : string
    strip_basenames: if True, then the basenames are removed before the split
    raiseError: if True
    Returns
    -------
    (commonpath,not_common)

    """
    if strip_basenames:
        path1 = os.path.basename(path1)
        path2 = os.path.basename(path2)

    commonpath = os.path.commonpath([path1,path2])
    if len(commonpath)==0:
        return('',path1)
    not_common = path1[len(commonpath):]
    not_common = not_common.lstrip("/")
    commonpath = commonpath.rstrip("/")
    return(commonpath,not_common)

def makepath(path,raiseError=True):
    if path == '':
        return(0)
    if not os.path.isdir(path):
        os.makedirs(path)
        if not os.path.isdir(path):
            if raiseError:
                raise RuntimeError('ERROR: Cannot create directory %s' % path)
            else:
                return(1)
    return(0)

def makepath4file(filename,raiseError=True):
    path = os.path.dirname(filename)
    if not os.path.isdir(path):
        return(makepath(path,raiseError=raiseError))
    else:
        return(0)

def rmfile(filename,raiseError=1,gzip=False):
    " if file exists, remove it "
    if os.path.lexists(filename):
        os.remove(filename)
        if os.path.isfile(filename):
            if raiseError == 1:
                raise RuntimeError('ERROR: Cannot remove %s' % filename)
            else:
                return(1)
    if gzip and os.path.lexists(filename+'.gz'):
        os.remove(filename+'.gz')
        if os.path.isfile(filename+'.gz'):
            if raiseError == 1:
                raise RuntimeError('ERROR: Cannot remove %s' % filename+'.gz')
            else:
                return(2)
    return(0)

def rmfiles(filenames,raiseError=1,gzip=False):
    if not (type(filenames) is list):
        raise RuntimeError("List type expected as input to rmfiles")
    errorflag = 0
    for filename in filenames:
        errorflag |= rmfile(filename,raiseError=raiseError,gzip=gzip)
    return(errorflag)


#https://numpy.org/doc/stable/reference/routines.set.html
def AorB(A,B):
    if len(A) == 0:
        return(B)
    if len(B) == 0:
        return(A)
    return(np.union1d(A,B))

def AandB(A,B,assume_unique=False,keeporder=False):
    if keeporder:
        # This is slower, but keeps order
        out=[]
        for i in A:
            if i in B:
                out.append(i)
        return(out)
    return(np.intersect1d(A,B,assume_unique=assume_unique))

def AnotB(A,B,keeporder=False):
    if keeporder:
        # This is slower, but keeps order
        out=[]
        for i in A:
            if not(i in B):
                out.append(i)
        return(out)
    return(np.setdiff1d(A,B))

def not_AandB(A,B):
    return(np.setxor1d(A,B))

def order_A_like_B(A,B):
    """
    Parameters
    ----------
    A : list-like object
    B : list-like object

    Returns
    -------
    returns A but ordered like the following:
        all entries that are also in B come first, in the order of B
        all entries that are not in B come next, in the original order

    """
    tmp = AandB(A,B,keeporder=True)
    notB = AnotB(A,tmp,keeporder=True)
    tmp.extend(notB)
    return(tmp)

def unique(A):
    unique = []
    for a in A:
        if a not in unique:
            unique.append(a)
    return unique

def radec2coord(ra, dec):
    unit = [u.deg, u.deg]
    if ':' in str(ra):
        # Assume input RA/DEC are hour/degree
        unit[0] = u.hourangle

    try:
        coord = SkyCoord(ra, dec, frame='icrs', unit=unit)
        return(coord)
    except ValueError:
        print(f'ERROR: cannot interpret: RA={ra} DEC={dec}')
        return(None)



class pdastroclass:
    def __init__(self,hexcols=[],hexcol_formatters={},**kwargs):
        self.t = pd.DataFrame(**kwargs)

        self.verbose = 0

        # if self.auto_convert_dtypes==True, then before .to_string() is called in self.write, self.t.convert_dtypes() is run
        #self.auto_convert_dtypes = True

        # special cases: columns in hexadecimal format.
        self.hexcols=hexcols
        self.hexcol_formatters=hexcol_formatters

        # if a file is successfully loaded, the filename is saved in this varviable
        self.filename = None


        # if self.default_dtypeMapping != None, then the dtype mapping is applied to the table .to_string() is called in self.write
        # This makes sure that formatters don't throw errors if the type of a column got changed to float or object during
        # one of the table operations
        self.default_dtypeMapping = None
        # example:
        # self.default_dtypeMapping = {'counter':np.int64}

        # default_formatters are passed to to_string() in self.write
        self.default_formatters = {}
        # example:
        # self.default_formatters = {'MJD':'{:.6f}'.format,'counter':'{:05d}'.format}

        # add list of columns to be skipped when using write function
        self.skipcols = []

        # dictionary for the splines. arguments are the y columns of the spline
        self.spline={}


    def load_lines(self,lines,sep='\s+',**kwargs):
        #errorflag = self.load_spacesep(io.StringIO('\n'.join(lines)),sep=sep,**kwargs)
        errorflag = self.load_spacesep(io.StringIO('\n'.join(lines)),**kwargs)
        return(errorflag)

    def load_cmpfile(self,filename,**kwargs):
        """
        load old frankenstein format of cmp file

        Parameters
        ----------
        filename :
            cmp filename.

        Returns
        -------
        errorflag and fits header

        """
        cmphdr = fits.getheader(filename)
        s = ''
        for i in range(1,int(cmphdr['NCOLTBL'])+1):
            s+=' %s' % cmphdr['COLTBL%d' % i]
        s = re.sub('Xpos','X',s)
        s = re.sub('Ypos','Y',s)
        print(s)
        lines = open(filename,'r').readlines()
        lines[0]=s
        errorflag = self.load_spacesep(io.StringIO('\n'.join(lines)))
        return(errorflag,cmphdr)

    def load_spacesep(self,filename,test4commentedheader=True,namesMapping=None,roundingMapping=None,
                      hexcols=None,auto_find_hexcols=True, delim_whitespace=True,
                      na_values=['None','-','--'],verbose=False,**kwargs):

        #kwargs['delim_whitespace']=True

        #also test for commented header to make it compatible to old format.
        self.load(filename,na_values=na_values,test4commentedheader=test4commentedheader,
                  namesMapping=namesMapping,roundingMapping=roundingMapping,delim_whitespace=delim_whitespace,
                  hexcols=hexcols,auto_find_hexcols=auto_find_hexcols,verbose=verbose,**kwargs)

        return(0)

    def load(self,filename,raiseError=True,test4commentedheader=False,namesMapping=None,roundingMapping=None,
             hexcols=None,auto_find_hexcols=True,verbose=False,delim_whitespace=True,**kwargs):
        #self.t = ascii.read(filename,format='commented_header',delimiter='\s',fill_values=[('-',0),('--',0)])

        try:
            if verbose: print('Loading %s' % filename)
            self.t = pd.read_table(filename,delim_whitespace=delim_whitespace,**kwargs)
            self.filename = filename
        except Exception as e:
            print('ERROR: could not read %s!' % filename)
            if raiseError:
                raise RuntimeError(str(e))
            self.filename = None
            return(1)

        if test4commentedheader:
            # This is to make it compatible to my old-style commented header files!
            # commented header: make sure it doesn't count the '#' as a column!
            if self.t.columns[0]=='#':
                renamemapping = {}
                for i in range(len(self.t.columns)-1):
                    renamemapping[self.t.columns[i]]=self.t.columns[i+1]
                renamemapping[self.t.columns[-1]]='__delme'
                self.t = self.t.rename(columns=renamemapping)
                self.t = self.t.drop(columns=['__delme'])

            # make sure the # is not kept in column name!
            if self.t.columns[0][0]=='#':
                self.t = self.t.rename(columns={self.t.columns[0]:self.t.columns[0][1:]})

        if hexcols is None: hexcols=[]
        hexcols.extend(self.hexcols)
        self.formattable(namesMapping=namesMapping,roundingMapping=roundingMapping,hexcols=hexcols,auto_find_hexcols=auto_find_hexcols)
        return(0)

    def write(self,filename=None,indices=None,columns=None,formatters=None,
              raiseError=True,overwrite=True,verbose=False,
              commentedheader=False,
              index=False, makepathFlag=True,convert_dtypes=False,
              hexcols=None,skipcols=None,
              return_lines=False,
              htmlflag=False, htmlsortedtable=False, **kwargs):

        # make sure indices are converted into a valid list
        indices=self.getindices(indices)

        # make sure columns are converted into a valid list
        columns=self.getcolnames(columns)
        if skipcols is None: skipcols = self.skipcols
        if len(skipcols)>0:
            columns=AnotB(columns,skipcols,keeporder=True)

        # make the path to the file if necessary
        if not (filename is None):
            if makepathFlag:
                if makepath4file(filename,raiseError=False):
                    errorstring='ERROR: could not make directory for %s' % filename
                    if raiseError:
                        raise RuntimeError(errorstring)
                    #print(errorstring)
                    return(1)

            # if overwrite, then remove the old file first
            if os.path.isfile(filename):
                if overwrite:
                    os.remove(filename)
                    if os.path.isfile(filename):
                        errorstring='ERROR: could not clobber %s' % filename
                        if raiseError:
                            raise RuntimeError(errorstring)
                        #print(errorstring)
                        return(2)
                else:
                    if raiseError:
                        raise RuntimeError(f'file {filename} already exists! use overwrite option...')

                    print(f'Warning: file {filename} already exists, not deleting it, skipping! if you want to overwrite, use overwrite option!')
                    return(0)

        # Fix the dtypes if wanted. THIS IS DANGEROUS!!!
        if convert_dtypes:
            # Using pandas convert_dtypes converts the columns into panda types, e.g. Int64. These types
            # are different than the numpy types (e.g., int64). The numpy types work with '{}'.format, the panda
            # int types do not!!! Therefore I set convert_dtype to False by default.
            self.t=self.t.convert_dtypes()
        if not(self.default_dtypeMapping is None):
            self.formattable(dtypeMapping=self.default_dtypeMapping)

        # if both formatters and self.defaultformatters are None, then no formatting. formatters supersedes self.defaultformatters
        if formatters is None:
            formatters = self.default_formatters

        # hexcols
        if hexcols is None: hexcols=[]
        hexcols.extend(self.hexcols)
        if len(hexcols)>0:
            if formatters is None: formatters={}
            for hexcol in self.hexcols:
                if not(hexcol in self.t.columns):
                    #Nothing to do yet!
                    continue
                if not (hexcol in formatters):
                    if isinstance(self.t[hexcol].dtype,np.int64):
                        formatters[hexcol]='0x{:08x}'.format
                    else:
                        formatters[hexcol]='0x{:04x}'.format

        if verbose>1 and not(filename is None): print('Saving %d rows into %s' % (len(indices),filename))

        # ugly hack: backwards compatibility to old commented header texttable format:
        # rename first column temporarily to have a '#'
        if commentedheader:
            columns=list(columns)
            self.t.rename(columns={columns[0]:f'#{columns[0]}'},inplace=True)
            columns[0]=f'#{columns[0]}'

        if len(indices)==0:
            if columns is None: columns = []
            # just save the header
            lines = [' '.join(columns)+'\n']
        else:
            if htmlflag:
                lines = self.t.loc[indices].to_html(index=index, columns=columns, formatters=formatters, **kwargs)
                if htmlsortedtable:
                    lines = re.sub('^\<table','<script type="text/javascript" src="sortable.js"></script>\n<table class="sortable" id="anyid" ',lines)
            else:
                lines = self.t.loc[indices].to_string(index=index, columns=columns, formatters=formatters, **kwargs)
        if filename is None:
            if not return_lines:
                print(lines)
        else:
            open(filename,'w').writelines(lines)


            """
            if filename is None:
                if return_lines:
                    return(0,self.t.loc[indices].to_string(index=index, columns=columns, formatters=formatters, **kwargs))
                else:
                    print(self.t.loc[indices].to_string(index=index, columns=columns, formatters=formatters, **kwargs))
            else:
                if not htmlflag:
                    self.t.loc[indices].to_string(filename, index=index, columns=columns, formatters=formatters, **kwargs)
                else:
                    if htmlsortedtable:
                        lines = self.t.loc[indices].to_html(index=index, columns=columns, formatters=formatters, **kwargs)
                        lines = re.sub('^\<table','<script type="text/javascript" src="sortable.js"></script>\n<table class="sortable" id="anyid" ',lines)
                        f = open(filename,"w")
                        f.writelines(lines)
                        f.close()
                    else:
                        self.t.loc[indices].to_html(filename, index=index, columns=columns, formatters=formatters, **kwargs)
                    #self.t.loc[indices].to_html(filename, index=index, columns=columns, formatters=formatters, **kwargs)
            """

        # reverse ugly hack
        if commentedheader:
            columns[0]=re.sub('^\#','',columns[0])
            self.t.rename(columns={"#"+columns[0]:columns[0]},inplace=True)

        if filename is not None:
            # some extra error checking...
            if not os.path.isfile(filename):
                errorstring='ERROR: could not save %s' % filename
                if raiseError:
                    raise RuntimeError(errorstring)
                #print(errorstring)
                return(3)
        if return_lines:
            return(0,lines)
        else:
            return(0)

    def formattable(self,namesMapping=None,roundingMapping=None,dtypeMapping=None,hexcols=None,auto_find_hexcols=False):

        if not(namesMapping is None):
            self.t = self.t.rename(columns=namesMapping)

        if not(roundingMapping is None):
            self.t = self.t.round(roundingMapping)

        if not(dtypeMapping is None):
            for col in dtypeMapping:
                self.t[col] = self.t[col].astype(dtypeMapping[col])


        # hexadecimal columns
        hexpattern = re.compile('^0x[a-fA-F0-9]+$')
        # if wanted, find columns that are in hexadecimal string format, and convert it interger.
        # These columns are added to self.hexcols
        if auto_find_hexcols and len(self.t)>0:
            for col in self.t.columns:
                if self.t[col].dtype == object and isinstance(self.t.at[0,col],str):
                    if not(hexpattern.search(self.t.at[0,col]) is None):
                        # add column to self.hexcols, so that it stays a hexcol when writing and saving
                        if not(col in self.hexcols):
                            self.hexcols.append(col)
                        # convert it to int
                        self.t[col] = self.t[col].apply(int, base=16)
        # Go through hexcols, and check if they need conversion.
        # These columns are also added to self.hexcols
        if not(hexcols is None):
            for hexcol in hexcols:
                # add column to self.hexcols, so that it stays a hexcol when writing and saving
                if not(hexcol in self.hexcols):
                    self.hexcols.append(hexcol)
                if not(hexcol in self.t.columns):
                    # nothing to do (yet)!
                    continue
                # if the column is still a string starting with '0x', convert it to int
                if self.t[hexcol].dtype == object and isinstance(self.t.at[0,hexcol],str):
                    if not(hexpattern.search(self.t.at[0,hexcol]) is None):
                         self.t[hexcol] = self.t[hexcol].apply(int, base=16)
        return(0)

    def getindices(self,indices=None):
        """make indices conform (input can be None,([list],), int, str, or list). The output is a list """

        #If indices is None, return all values
        if indices is None:
            return(self.t.index.values)

        # If indices=([indiceslist],), then it needs to be changed to indices=indiceslist
        if isinstance(indices,tuple):
            if len(indices)==0:
                #return an empty list instead of empty tuple
                return([])
            # teh first entry is a list, then these are the relevant indices!
            if isinstance(indices[0],list) or isinstance(indices[0],np.ndarray):
                return(indices[0])
            else:
                return(list(indices))

        # if the passed value is an integer or str, make it a list!
        if isinstance(indices,int) or isinstance(indices,str) or isinstance(indices,float):
            return([indices])

        indices=np.array(indices)

        #if not (isinstance(indices,list) or isinstance(indices,np.ndarray)):
        #    raise RuntimeError("Can't convert this to an indices list!",type(indices),indices)

        return(indices)

    def getcolnames(self,colnames=None):
        """Return a list of all colnames of colnames=None. If colnames=string, return a list"""
        if (colnames is None):
             colnames = self.t.columns[:]
        elif isinstance(colnames,str):
            if colnames.lower()=='all':
                colnames = self.t.columns[:]
            else:
                colnames=[colnames]
        return(colnames)

    def ix_remove_null(self,colnames=None,indices=None):
        print('ix_remove_null deprecated, replace with ix_not_null')
        return(self.ix_not_null(colnames=colnames,indices=indices))

    def ix_not_null(self,colnames=None,indices=None):
        # get the indices based on input.
        indices=self.getindices(indices)

        # get the column names over which to iterate
        colnames=self.getcolnames(colnames)

        for colname in colnames:
            #print('XXX',indices)
            (notnull,) = np.where(pd.notnull(self.t.loc[indices,colname]))
            indices = indices[notnull]
            #print('YYY',notnull)
        return(indices)

    def ix_is_null(self,colnames=None,indices=None):
        # get the indices based on input.
        indices=self.getindices(indices)

        # get the column names over which to iterate
        colnames=self.getcolnames(colnames)

        for colname in colnames:
            #print('XXX',indices)
            (null,) = np.where(pd.isnull(self.t.loc[indices,colname]))
            indices = indices[null]
            #print('YYY',null)
        return(indices)

    def ix_equal(self,colnames,val,indices=None):
        # get the indices based on input.

        # use isnull() if val is None
        if val is None:
            indices = self.ix_is_null(colnames,indices=indices)
            return(indices)

        indices=self.getindices(indices)

        # get the column names over which to iterate
        colnames=self.getcolnames(colnames)
        for colname in colnames:
            (keep,) = np.where(self.t.loc[indices,colname].eq(val))
            indices = indices[keep]

        return(indices)

    def ix_inrange(self,colnames=None,lowlim=None,uplim=None,indices=None,
                   exclude_lowlim=False,exclude_uplim=False):

        # get the indices based on input.
        indices=self.getindices(indices)

        # get the column names over which to iterate
        colnames=self.getcolnames(colnames)
        #print(colnames)
        for colname in colnames:
            if not(lowlim is None):
                if exclude_lowlim:
                    (keep,) = np.where(self.t.loc[indices,colname].gt(lowlim))
                else:
                    (keep,) = np.where(self.t.loc[indices,colname].ge(lowlim))
                indices = indices[keep]
                #print('lowlim cut:',keep)

            if not(uplim is None):
                if exclude_uplim:
                    (keep,) = np.where(self.t.loc[indices,colname].lt(uplim))
                else:
                    (keep,) = np.where(self.t.loc[indices,colname].le(uplim))
                indices = indices[keep]
                #print('uplim cut:',keep)
        return(indices)

    def ix_outrange(self,colnames=None,lowlim=None,uplim=None,indices=None,
                    exclude_lowlim=False,exclude_uplim=False):

        # get the indices based on input.
        indices=self.getindices(indices)

        # get the column names over which to iterate
        colnames=self.getcolnames(colnames)

        #print('BBB',indices)
        for colname in colnames:
            if not(lowlim is None):
                if exclude_lowlim:
                    (keeplow,) = np.where(self.t.loc[indices,colname].lt(lowlim))
                else:
                    (keeplow,) = np.where(self.t.loc[indices,colname].le(lowlim))
                #print('lowlim cut:',keeplow)
            else:
                keeplow=[]

            if not(uplim is None):
                if exclude_uplim:
                    (keepup,) = np.where(self.t.loc[indices,colname].gt(uplim))
                else:
                    (keepup,) = np.where(self.t.loc[indices,colname].ge(uplim))
                #print('uplim cut:',keepup)
            else:
                keepup=[]

            indices = indices[AorB(keeplow,keepup)]

        return(indices)

    def ix_unmasked(self,maskcol,maskval=None,indices=None):

        # get the indices based on input.
        indices=self.getindices(indices)

        if maskval is None:
            (keep,) = np.where(self.t.loc[indices,maskcol].eq(0))
        else:
            (keep,) = np.where(bitmask.bitfield_to_boolean_mask(self.t.loc[indices,maskcol].astype('int'),ignore_flags=~maskval,good_mask_value=True))
        indices = indices[keep]
        return(indices)

    def ix_masked(self,maskcol,maskval=None,indices=None):

        # get the indices based on input.
        indices=self.getindices(indices)

        if maskval is None:
            (keep,) = np.where(self.t.loc[indices,maskcol].ne(0))
        else:
            (keep,) = np.where(bitmask.bitfield_to_boolean_mask(self.t.loc[indices,maskcol].astype('int'),ignore_flags=~maskval))
        indices = indices[keep]
        return(indices)

    def ix_matchregex(self,col,regex,indices=None):
        # get the indices based on input.
        indices=self.getindices(indices)

        (keep,) = np.where((self.t.loc[indices,col].str.contains(regex)==True))
        #bla = self.t[col].str.contains(regex)==True

        indices = indices[keep]
        return(indices)

    def ix_sort_by_cols(self,cols,indices=None):

        # get the indices based on input.
        indices=self.getindices(indices)

        # get the column names (makes sure that it is a list)
        cols=self.getcolnames(cols)

        ix_sorted = self.t.loc[indices].sort_values(cols).index.values

        return(ix_sorted)

    def newrow(self,dicti=None):
        row_df = pd.DataFrame(dicti, index=[len(self.t)])
        self.t = pd.concat([self.t, row_df])
        return(self.t.index.values[-1])

    def add2row(self,index,dicti):
        self.t.loc[index,list(dicti.keys())]=list(dicti.values())
        return(index)

    def fitsheader2table(self,fitsfilecolname,indices=None,requiredfitskeys=None,optionalfitskey=None,
                         raiseError=True,verify='silentfix',skipcolname=None,headercol=None,ext=None,extname=None,
                         prefix=None,suffix=None):
        def fitskey2col(fitskey,prefix=None,suffix=None):
            col = fitskey
            if prefix is not None: col = prefix+col
            if suffix is not None: col += suffix
            return(col)


        indices = self.getindices(indices)

        # initialize columns if necessary
        if requiredfitskeys!=None:
            for fitskey in requiredfitskeys:
                col = fitskey2col(fitskey,prefix=prefix,suffix=suffix)
                if not (col in self.t.columns):
                    self.t[col]=None
        if optionalfitskey!=None:
            for fitskey in optionalfitskey:
                col = fitskey2col(fitskey,prefix=prefix,suffix=suffix)
                if not (col in self.t.columns):
                    self.t[col]=None

        if headercol!=None and (not (headercol in self.t.columns)):
            self.t[headercol]=None

        # loop through the images
        for index in indices:
            #header = fits.getheader(self.t.loc[index,fitsfilecolname],ext=ext,extname=extname)
            # It was impossible to use 'verify' with getheader...
            hdu = fits.open(self.t.loc[index,fitsfilecolname],ext=ext,extname=extname,output_verify="silentfix")
            if verify is not None: hdu.verify(verify)
            header = hdu[0].header
            if headercol!=None:
                self.t[headercol]=header

            if skipcolname!=None:
                self.t.loc[index,skipcolname]=False
            if requiredfitskeys!=None:
                for fitskey in requiredfitskeys:
                    col = fitskey2col(fitskey,prefix=prefix,suffix=suffix)
                    if fitskey in header:
                        self.t.loc[index,col]=header[fitskey]
                    else:
                        if raiseError:
                            raise RuntimeError("fits key %s does not exist in file %s" % (fitskey,self.t[fitsfilecolname][index]))
                        else:
                            self.t.loc[index,col]=None
                            if skipcolname!=None:
                                 self.t.loc[index,skipcolname]=True

            if optionalfitskey!=None:
                for fitskey in optionalfitskey:
                    col = fitskey2col(fitskey,prefix=prefix,suffix=suffix)
                    if fitskey in header:
                        self.t.loc[index,col]=header[fitskey]
                    else:
                        self.t.loc[index,col]=None

    def dateobs2mjd(self,dateobscol,mjdcol,timeobscol=None,indices=None,tformat='isot'):
        indices = self.getindices(indices)
        if len(indices)==0:
            return(0)

        if not (mjdcol in self.t.columns):
            self.t.loc[indices,mjdcol]=None

        if timeobscol!=None:
            dateobslist = list(self.t.loc[indices,dateobscol]+'T'+self.t.loc[indices,timeobscol])
        else:
            dateobslist = list(self.t.loc[indices,dateobscol])

        dateobjects = Time(dateobslist,  format=tformat, scale='utc')
        mjds = dateobjects.mjd
        self.t.loc[indices,mjdcol]=mjds

    def mjd2dateobs(self,mjdcol,dateobscol,indices=None,tformat='isot'):
        indices = self.getindices(indices)
        if len(indices)==0:
            return(0)

        if not (dateobscol in self.t.columns):
            self.t.loc[indices,dateobscol]=None

        mjdlist = list(self.t.loc[indices,mjdcol])

        dateobjects = Time(mjdlist, format='mjd')
        dateobs = dateobjects.to_value('isot')
        self.t.loc[indices,dateobscol]=dateobs

    def calc_color(self,f1,df1,f2,df2,outcolor=None,outcolor_err_nameformat='e(%s)',indices=None,color_formatter='{:.3f}'.format):
        indices = self.getindices(indices)
        if len(indices)==0:
            return(0)

        # get color and uncertainty column names
        if outcolor is None: outcolor = '%s-%s' % (f1,f2)
        outcolor_err = outcolor_err_nameformat % outcolor
        # delete all previous results
        self.t.loc[indices,outcolor]= np.nan
        self.t.loc[indices,outcolor_err]= np.nan

        # Only use indices for which both filters have uncertainties, i.e. they are not upper limits
        ix_good = self.ix_remove_null(df1,indices=indices)
        ix_good = self.ix_remove_null(df2,indices=ix_good)

        self.t[outcolor] = self.t.loc[ix_good,f1] - self.t.loc[ix_good,f2]
        self.t[outcolor_err] = np.sqrt(np.square(self.t.loc[ix_good,df1]) + np.square(self.t.loc[ix_good,df2]))

        if not(color_formatter is None):
            if self.default_formatters is None:
                self.default_formatters = {}
            self.default_formatters[outcolor]=color_formatter
            self.default_formatters[outcolor_err]=color_formatter

        return(0)

    def radeccols_to_SkyCoord(self,racol=None,deccol=None,indices=None, frame='icrs',
                              assert_0_360_limits=True,assert_pm90_limits=True):
        if (racol is None) and (deccol is None):
            raise RuntimeError('You need to specify at least one of racol or deccol')

        indices = self.getindices(indices)
        for col in [racol,deccol]:
            if col is None: continue
            ixs_null = self.ix_is_null(col,indices)
            if len(ixs_null)>0:
                self.write(indices=ixs_null)
                raise RuntimeError(f'Null values for column {col} in above row(s)')

        if len(indices)==0:
            print('Warning, trying to assert ra/dec columns are decimal for 0 rows')
            return([],[])

        # fill ra and dec
        ra = np.full(len(indices),np.nan)
        dec = np.full(len(indices),np.nan)
        if racol is not None: ra = np.array(self.t.loc[indices,racol])
        if deccol is not None: dec = np.array(self.t.loc[indices,deccol])

        # check if all cols are already in numerical format. If yes, all good! This can speed things up!!
        numflag=True
        for col in [racol,deccol]:
            if col is None: continue
            if not(col in self.t.columns):
                raise RuntimeError(f'Column {col} is not in columns {self.t.columns}')
            if not(is_float_dtype(self.t[col]) or is_integer_dtype(self.t[col])):
                numflag=False

        if numflag:
            # no conversion needed!
            # check ra dec limits if wanted
            if (ra is not None) and assert_0_360_limits:
                ra = np.where(ra<0.0,ra+360.0,ra)
                ra = np.where(ra>=360.0,ra-360.0,ra)
            if (dec is not None) and assert_pm90_limits:
                dec = np.where(dec<-90.0,dec+180.0,dec)
                dec = np.where(dec> 90.0,dec-180.0,dec)

            coord = SkyCoord(ra, dec, frame=frame, unit=(u.deg,u.deg))
        else:
            # convert the strings into decimal degrees
            unit = [u.deg,u.deg]

            check_line_by_line = False
            # check format of racol if necessary
            if racol is not None:
                hexagesimal = unique([(re.search(':',x) is not None) for x in ra])
                if len(hexagesimal)==1:
                    # if hexagesimal true, set unit of RA to hourangle!!
                    if hexagesimal[0]:
                        unit[0] = u.hourangle
                elif len(hexagesimal)==2:
                    # both formats? then check line by line!
                    print('Warning: it looks like there are inconsistent formats in RA column {racol}! checking line by line for sexagesimal')
                    check_line_by_line = True
                else:
                    raise RuntimeError('Something is wrong here when trying to determine if RA col {racol} is sexagesimal! ')


            if check_line_by_line:
                coord = np.full(len(indices),np.nan)
                hexagesimal = [(re.search(':',x) is not None) for x in ra]
                raunits=np.where(hexagesimal,u.hourangle,u.deg)
                for i in range(len(ra)):
                    coord[i] = SkyCoord(ra[i], dec[i], frame=frame, unit=(raunits[i],u.deg))
            else:
                coord = SkyCoord(ra, dec, frame=frame, unit=unit)
            # no need to check ra dec limits, already done in SkyCoord

        return(indices,coord)


    def assert_radec_cols_decimal_degrees(self,racol=None,deccol=None,
                                          outracol=None,outdeccol=None,
                                          indices=None,coordcol=None,
                                          assert_0_360_limits=True,
                                          assert_pm90_limits=True):

        (indices,coord) = self.radeccols_to_SkyCoord(racol=racol,deccol=deccol,indices=indices,
                                                     assert_0_360_limits=assert_0_360_limits,assert_pm90_limits=assert_pm90_limits)

        if outracol is None: outracol = racol
        if outdeccol is None: outdeccol = deccol
        if outracol is not None: self.t.loc[indices,outracol]=coord.ra.degree
        if outdeccol is not None: self.t.loc[indices,outdeccol]=coord.dec.degree
        if coordcol is not None:
            self.t.loc[indices,coordcol]=coord
            # Don't write coordcol
            if not (coordcol in self.skipcols): self.skipcols.append(coordcol)

        return(0)

    def assert_radec_cols_sexagesimal(self,racol=None,deccol=None,
                                      outracol=None,outdeccol=None,
                                      indices=None,coordcol=None,
                                      precision=3,
                                      assert_0_360_limits=True,
                                      assert_pm90_limits=True):

        (indices,coord) = self.radeccols_to_SkyCoord(racol=racol,deccol=deccol,indices=indices,
                                                     assert_0_360_limits=assert_0_360_limits,
                                                     assert_pm90_limits=assert_pm90_limits)

        if outracol is None: outracol = racol
        if outdeccol is None: outdeccol = deccol

        # list of ra/dec pairs
        hmsdms = map(lambda x: x.split(' '), coord.to_string(sep=':', style='hmsdms', precision=precision))
        # unzip pairs into a list of ra and a list of dec
        radeclist = list(zip(*hmsdms))

        if outracol is not None: self.t.loc[indices,outracol]=radeclist[0]
        if outdeccol is not None: self.t.loc[indices,outdeccol]=radeclist[1]
        if coordcol is not None:
            self.t.loc[indices,coordcol]=coord
            # Don't write coordcol
            if not (coordcol in self.skipcols): self.skipcols.append(coordcol)

        return(0)


    def radeccols_to_coord(self,racol,deccol,coordcol,indices=None,
                           assert_0_360_limits=True,assert_pm90_limits=True):

        (indices,coord) = self.radeccols_to_SkyCoord(racol=racol,deccol=deccol,indices=indices,
                                                     assert_0_360_limits=assert_0_360_limits,assert_pm90_limits=assert_pm90_limits)

        self.t.loc[indices,coordcol]=coord
        # Don't write coordcol
        if not (coordcol in self.skipcols): self.skipcols.append(coordcol)

        return(0)


    def flux2mag(self,fluxcol,dfluxcol,magcol,dmagcol,indices=None,
                 zpt=None,zptcol=None, upperlim_Nsigma=None):

        indices = self.getindices(indices)
        if len(indices)==0:
            return(0)

        # delete all previous results
        self.t.loc[indices,magcol]= np.nan
        self.t.loc[indices,dmagcol]= np.nan

        # if upperlim_Nsigma is not None, then upper limits are calculated.
        if upperlim_Nsigma is None:
            indices_mag = indices
            # no upper limits
            indices_ul = indices_ul_negative = []
        else:
            # calculate the S/N
            self.t.loc[indices,'__tmp_SN']=self.t.loc[indices,fluxcol]/self.t.loc[indices,dfluxcol]
            # Get the indices with valid S/N
            indices_validSN = self.ix_not_null('__tmp_SN',indices=indices)
            # get the indices for which the S/N>=upperlim_Nsigma
            indices_mag = self.ix_inrange(['__tmp_SN'],upperlim_Nsigma,indices=indices_validSN)
            # all the other indices can only be used for upper limits
            indices_ul = AnotB(indices_validSN,indices_mag)
            # get the indices for which the S/N is negative
            indices_ul_negative = self.ix_inrange(['__tmp_SN'],None,0.0,indices=indices_ul)
            #self.t = self.t.drop(columns=['__tmp_SN'])

        # Calculate the mags and uncertainties
        self.t.loc[indices_mag,magcol]= -2.5*np.log10(self.t.loc[indices_mag,fluxcol])
        self.t.loc[indices_mag,dmagcol] = 2.5 / np.log(10.0) * self.t.loc[indices_mag,dfluxcol]/self.t.loc[indices_mag,fluxcol]

        # Are there upper limits to be calculated?
        if len(indices_ul)>0:
            # just to be sure, set dmags to nan
            self.t.loc[indices_ul,dmagcol] = np.nan

            indices_ul_positive = AnotB(indices_ul,indices_ul_negative)
            if len(indices_ul_positive)>0:
                # If flux is positive: upper limit = flux + Nsigma * dflux
                self.t.loc[indices_ul_positive,magcol] = -2.5*np.log10(self.t.loc[indices_ul_positive,fluxcol]+upperlim_Nsigma*self.t.loc[indices_ul_positive,dfluxcol])

            if len(indices_ul_negative)>0:
                # If flux is negative: upper limit = Nsigma * dflux
                self.t.loc[indices_ul_negative,magcol] = -2.5*np.log10(upperlim_Nsigma*self.t.loc[indices_ul_negative,dfluxcol])

        if not(zpt is None):
            self.t.loc[indices,magcol]+=zpt
        if not(zptcol is None):
            self.t.loc[indices,magcol]+=self.t.loc[indices,zptcol]


    def initspline(self,xcol,ycol,indices=None,
                   kind='cubic',bounds_error=False,fill_value='extrapolate',
                   **kwargs):
        if not (xcol in self.t.columns):
            raise RuntimeError("spline: x column %s does not exist in table!" % xcol)
        if not (ycol in self.t.columns):
            raise RuntimeError("spline: y column %s does not exist in table!" % ycol)

        # make sure there are no nan values
        indices = self.ix_remove_null(colnames=[xcol,ycol])

        # initialize the spline and save it in self.spline with the key ycol
        self.spline[ycol]= interp1d(self.t.loc[indices,xcol],self.t.loc[indices,ycol],
                                    kind=kind,bounds_error=bounds_error,fill_value=fill_value,**kwargs)

    def getspline(self,xval,ycol):
        if not(ycol in self.spline):
            raise RuntimeError('Spline for column %s is not defined!' % ycol)
        return(self.spline[ycol](xval))

    def plot(self,indices=None,*args,**kwargs):
        indices = self.getindices(indices)
        ax = self.t.loc[indices].plot(*args,**kwargs)
        return(ax)


class pdastrostatsclass(pdastroclass):
    def __init__(self,**kwargs):
        pdastroclass.__init__(self,**kwargs)
        self.reset()
        self.set_statstring_format()
        self.c4_smalln = [0.0, 0.0, 0.7978845608028654, 0.8862269254527579, 0.9213177319235613, 0.9399856029866251, 0.9515328619481445]

    def c4(self,n):
        #http://en.wikipedia.org/wiki/Unbiased_estimation_of_standard_deviation
        if n<=6:
            return(self.c4_smalln[n])
        else:
            return(1.0 - 1.0/(4.0*n) - 7.0/(32.0*n*n) - 19.0/(128.0*n*n*n))

    def reset(self):
        self.statparams = {}
        for k  in ['mean','mean_err','stdev','stdev_err','X2norm','ix_good','ix_clip']:
            self.statparams[k]=None
        for k  in ['Ngood','Nclip','Nchanged','Nmask','Nnan','converged','i']:
            self.statparams[k]=0
        self.calc_stdev_X2_flag = True

    def set_statstring_format(self,format_floats='{:f}',format_ints='{:d}',format_none='{}', format_X2norm='{:.2f}'):
        self.str_format1 = "mean:%s(%s) stdev:%s(%s) X2norm:%s Nchanged:%s Ngood:%s Nclip:%s" % (format_floats,format_floats,format_floats,format_floats,format_X2norm,format_ints,format_ints,format_ints)
        self.str_format_none = "mean:%s(%s) stdev:%s(%s) X2norm:%s Nchanged:%s Ngood:%s Nclip:%s" % (format_none,format_none,format_none,format_none,format_none,format_none,format_none,format_none)
        #self.str_format2 = "mean:%s(%s) stdev:%s X2norm:%s Nchanged:%s Ngood:%s Nclip:%s" % (format_floats,format_floats,format_floats,format_floats,format_ints,format_ints,format_ints)

    def statstring(self):
        if self.statparams['mean'] is None or self.statparams['stdev'] is None:
            formatstring = "WARNING! i:{:02d} "+ self.str_format_none
        else:
            formatstring = "i:{:02d} "+ self.str_format1

        s = formatstring.format(self.statparams['i'],self.statparams['mean'],self.statparams['mean_err'],
                                self.statparams['stdev'],self.statparams['stdev_err'],self.statparams['X2norm'],
                                self.statparams['Nchanged'],self.statparams['Ngood'],self.statparams['Nclip'])
        return(s)



    def calcaverage_errorcut(self,datacol, noisecol, indices=None,
                             mean=None,Nsigma=None,medianflag=False,
                             return_ix=False, verbose=0):

        # get the indices based on input.
        indices=self.getindices(indices)

        # If N-sigma cut and second iteration (i.e. we have a stdev from the first iteration), skip bad measurements.
        if not(Nsigma is None) and not(mean is None):
            ix_good_bkp = copy.deepcopy(self.statparams['ix_good'])
            (keep,) = np.where(np.absolute(self.t.loc[indices,datacol]-mean)<=Nsigma*self.t.loc[indices,noisecol])
            ix_good  = indices[keep]
        else:
            ix_good_bkp = None
            ix_good  = indices

        if verbose>3 and not(ix_good_bkp is None) and not(ix_good is None):
            print('{} good data after sigma clipping, {} clipped'.format(len(ix_good),len(indices)-len(ix_good)))
            self.write(indices=ix_good)
            #self.write(columns=['objID','filter','psfFlux','psfFluxErr','psfMag','psfMagErr','psfMagErr_tot'],indices=ix_good)

        Ngood = len(ix_good)

        if Ngood>1:
            if medianflag:
                mean = self.t.loc[ix_good,datacol].median()
                if verbose>1: print('median: {:f}'.format(mean))
                stdev =  np.sqrt(1.0/(Ngood-1.0)*np.sum(np.square(self.t.loc[ix_good,datacol] - mean)))/self.c4(Ngood)
                mean_err = self.t.loc[ix_good,noisecol].median()/np.sqrt(Ngood-1)
                #mean_err = stdev/np.sqrt(Ngood-1)

            else:
                c1 = np.sum(1.0*self.t.loc[ix_good,datacol]/np.square(self.t.loc[ix_good,noisecol]))
                c2 = np.sum(1.0/np.square(self.t.loc[ix_good,noisecol]))
                mean = c1/c2
                mean_err = np.sqrt(1.0/c2)
                stdev = self.t.loc[ix_good,datacol].std()

            stdev_err = 1.0*stdev/np.sqrt(2.0*Ngood)
            X2norm = 1.0/(Ngood-1.0)*np.sum(np.square((self.t.loc[ix_good,datacol] - mean)/self.t.loc[ix_good,noisecol]))

        else:
            if Ngood==1:
                mean = self.t.loc[ix_good[0],datacol]*1.0
                mean_err = self.t.loc[ix_good[0],noisecol]*1.0
            else:
                mean = None
                mean_err = None

            X2norm   = None
            stdev     = None
            stdev_err = None

        self.statparams['ix_good']=ix_good
        self.statparams['Ngood']=Ngood
        self.statparams['ix_clip']=AnotB(indices,ix_good)
        self.statparams['Nclip']=len(indices) - Ngood
        if not(ix_good_bkp is None):
            self.statparams['Nchanged'] = len(not_AandB(ix_good_bkp,ix_good))
        else:
            self.statparams['Nchanged'] = 0

        self.statparams['mean']      = mean
        self.statparams['stdev']     = stdev
        self.statparams['mean_err']  = mean_err
        self.statparams['stdev_err'] = stdev_err
        self.statparams['X2norm']    = X2norm

        if Ngood<1:
            return(1)
        return(0)

    def calcaverage_sigmacut(self,datacol, noisecol=None, indices=None,
                             mean=None,stdev=None,Nsigma=None,
                             percentile_cut=None,percentile_Nmin=3,
                             medianflag=False,
                             return_ix=False, verbose=0,rescol='__tmp_residuals'):

        # get the indices based on input.
        indices=self.getindices(indices)
        if len(indices)==0:
            print('Warning!! no data passed for sigma cut!')
            self.reset()
            return(2)

        ix_good_bkp = None
        if (percentile_cut is None) or (len(indices)<=percentile_Nmin):
            # If N-sigma cut and second iteration (i.e. we have a stdev from the first iteration), skip bad measurements.
            if not(Nsigma is None) and not(stdev is None) and not(mean is None):
                ix_good_bkp = copy.deepcopy(self.statparams['ix_good'])
                (keep,) = np.where(np.absolute(self.t.loc[indices,datacol]-mean)<=Nsigma*stdev)
                ix_good  = indices[keep]
                if verbose>3:
                    print('good data after sigma clipping:')
                    self.write(indices=ix_good)
                    #self.write(columns=['objID','filter','psfFlux','psfFluxErr','psfMag','psfMagErr'],indices=ix_good)
            else:
                if verbose>3:
                    print('No sigma clipping yet...')
                ix_good  = indices
        else:
            # percentile clipping!!!
            if mean is None:
                if medianflag:
                    median = self.t.loc[indices,datacol].median()
                    if verbose>1: print('median: {:f}'.format(median))
                    mean = median
                else:
                    mean = self.t.loc[indices,datacol].mean()
                    if verbose>1: print('mean: {:f}'.format(mean))

            self.t[rescol]=np.absolute(self.t.loc[indices,datacol]-mean)
            max_residual = np.percentile(np.absolute(self.t.loc[indices,datacol]-mean),percentile_cut)
            if verbose:
                print('%f percentile cut: max residual for cut: %f' % (percentile_cut,max_residual))
            ix_good  = self.ix_inrange(rescol,None,max_residual,exclude_uplim=True)


            #print('all')
            #self.write(columns=['objID','psfMag','psfMagErr',rescol],indices=indices)
            #print('good')
            #self.write(columns=['objID','psfMag','psfMagErr',rescol],indices=ix_good)

            # make sure the percentile clipping does not cut too much away for small numbers: always keep the best percentile_Nmin
            if len(ix_good)<percentile_Nmin:
                print('Warning: %d<%d made it through the percentile cut, too few, taking the %d with the least residuals' % (len(ix_good),percentile_Nmin,percentile_Nmin))
                residuals = np.sort(self.t.loc[indices,rescol])
                max_residual = residuals[percentile_Nmin-1]
                ix_good  = self.ix_inrange(rescol,None,max_residual,exclude_uplim=False)
                if len(ix_good)<percentile_Nmin:
                    raise RuntimeError('%d<%d in percentile cut!' % (len(ix_good),percentile_Nmin))

            if verbose>3:
                print('good data after percentile clipping:')
                self.write(indices=ix_good)
                #self.write(columns=['objID','filter','psfFlux','psfFluxErr','psfMag','psfMagErr',rescol],indices=ix_good)


            #sys.exit(0)

            #residuals = np.sort(np.absolute(self.t.loc[indices,datacol]-mean))
            #print(residuals)
            #max_residual = np.percentile(np.absolute(self.t.loc[indices,datacol]-mean),percentile_cut)

            #sys.exit(0)


        Ngood = len(ix_good)
        if Ngood>1:
            if medianflag:
                median = self.t.loc[ix_good,datacol].median()
                #mean = scipy.median(self.t.loc[ix_good,datacol])
                if verbose>1: print('median: {:f}'.format(median))
                stdev =  np.sqrt(1.0/(Ngood-1.0)*np.sum(np.square(self.t.loc[ix_good,datacol] - median)))/self.c4(Ngood)
                mean = median
            else:
                mean = self.t.loc[ix_good,datacol].mean()
                if verbose>1: print('mean: {:f}'.format(mean))
                stdev = self.t.loc[ix_good,datacol].std()

            mean_err = stdev/np.sqrt(Ngood-1)
            stdev_err = 1.0*stdev/np.sqrt(2.0*Ngood)
            if noisecol is None:
                X2norm = 1.0/(Ngood-1.0)*np.sum(np.square((self.t.loc[ix_good,datacol] - mean)/stdev))
            else:
                X2norm = 1.0/(Ngood-1.0)*np.sum(np.square((self.t.loc[ix_good,datacol] - mean)/self.t.loc[ix_good,noisecol]))
        else:
            if Ngood==1:
                mean = self.t.loc[ix_good[0],datacol]*1.0
                if noisecol is None:
                    mean_err = None
                else:
                    mean_err = self.t.loc[ix_good[0],noisecol]*1.0
            else:
                mean = None
                mean_err = None

            X2norm   = None
            stdev     = None
            stdev_err = None

        self.statparams['ix_good']=ix_good
        self.statparams['Ngood']=Ngood
        self.statparams['ix_clip']=AnotB(indices,ix_good)
        self.statparams['Nclip']=len(indices) - Ngood
        if not(ix_good_bkp is None):
            self.statparams['Nchanged'] = len(not_AandB(ix_good_bkp,ix_good))
        else:
            self.statparams['Nchanged'] = 0

        self.statparams['mean']      = mean
        self.statparams['stdev']     = stdev
        self.statparams['mean_err']  = mean_err
        self.statparams['stdev_err'] = stdev_err
        self.statparams['X2norm']    = X2norm

        if Ngood<1:
            return(1)
        return(0)

    def calcaverage_sigmacutloop(self,datacol, indices=None, noisecol=None, sigmacutFlag=False,
                                 maskcol=None, maskval=None,
                                 removeNaNs = True,
                                 Nsigma=3.0,Nitmax=10,verbose=0,
                                 percentile_cut_firstiteration=None,
                                 median_firstiteration=True):
        """
         median_firstiteration: in the first iteration, use the median instead the mean. This is more robust if there is a population of bad measurements
        """

        if noisecol is None:
            sigmacutFlag = True

        # get the indices based on input.
        indices=self.getindices(indices)

        self.reset()

        # exclude data if wanted
        if maskcol!=None:
            Ntot = len(indices)
            indices = self.ix_unmasked(maskcol,maskval=maskval,indices=indices)
            self.statparams['Nmask']= Ntot-len(indices)
            if verbose>1: print('Keeping {:d} out of {:d}, skippin {:d} because of masking in column {} (maskval={})'.format(len(indices),Ntot,Ntot-len(indices),maskcol,maskval))
        else:
            self.statparams['Nmask']= 0

        # remove null values if wanted
        if removeNaNs:
            colnames = [datacol]
            if not(noisecol is None): colnames.append(noisecol)
            if not(maskcol is None): colnames.append(maskcol)
            Ntot = len(indices)
            indices = self.ix_remove_null(colnames,indices=indices)
            self.statparams['Nnan']= Ntot-len(indices)
            if verbose>1: print('Keeping {:d} out of {:d}, skippin {:d} because of null values in columns {:s}'.format(len(indices),Ntot,Ntot-len(indices),",".join(colnames)))
        else:
            self.statparams['Nnan']= 0


        while ((self.statparams['i']<Nitmax) or (Nitmax==0)) and (not self.statparams['converged']):
            # median only in first iteration and if wanted
            medianflag = median_firstiteration and (self.statparams['i']==0) and (Nsigma!=None)
            # percentile_cut only in first interation and if wanted
            percentile_cut = None
            if (self.statparams['i']==0):
                percentile_cut = percentile_cut_firstiteration

            if sigmacutFlag:
                errorflag = self.calcaverage_sigmacut(datacol, indices=indices, noisecol=noisecol,
                                                      mean = self.statparams['mean'], stdev = self.statparams['stdev'],
                                                      Nsigma=Nsigma,
                                                      medianflag=medianflag, percentile_cut=percentile_cut,
                                                      verbose=verbose)
            else:
                errorflag = self.calcaverage_errorcut(datacol, noisecol, indices=indices,
                                                      mean = self.statparams['mean'],
                                                      Nsigma=Nsigma, medianflag=medianflag, verbose=verbose)

            #if self.statparams['i']==0:
            #    self.statparams['stdev']=0.05

            if verbose>2:
                print(self.statstring())

            # Not converged???
#            if errorflag or self.statparams['stdev']==None or self.statparams['stdev']==0.0 or self.statparams['mean']==None:
            if errorflag or self.statparams['stdev']==None or (self.statparams['stdev']==0.0 and sigmacutFlag) or self.statparams['mean']==None:
                self.statparams['converged']=False
                break
            # Only do a sigma cut if wanted
            if Nsigma == None or Nsigma == 0.0:
                self.statparams['converged']=True
                break
            # No changes anymore? If yes converged!!!
            if (self.statparams['i']>0) and (self.statparams['Nchanged']==0) and (not medianflag):
                self.statparams['converged']=True
                break
            self.statparams['i']+=1
            print()

        if not(self.statparams['converged']):
            if self.verbose>1:
                print('WARNING! no convergence!')

        return(not self.statparams['converged'])

    """
    def colnames4params(self,columns=None,colmapping={},prefix='',suffix='',skipcols=[]):
        cols=[]
        keys=[]
        for k in ['mean','mean_err','stdev','stdev_err','X2norm','Ngood','Nclip','Nmask','Nnan','converged','i']:
            if k in skipcols:
                continue
            outcol=k
            if k in colmapping:
                outcol=colmapping[k]
            keys.append(k)
            cols.append('{}{}{}'.format(prefix,outcol,suffix))
        return(keys,cols)
"""


    def intializecols4statparams(self,params=None,prefix='',suffix='',parammapping={},skipparams=[],
                                 addformat2defaultformatter=True,format4outvals='{:.3f}',
                                 setcol2None=True):
        outcols=[]
        outparams=[]

        if params is None:
            params=['mean','mean_err','stdev','stdev_err','X2norm','Ngood','Nclip','Nmask','Nnan','converged','i']

        if addformat2defaultformatter and (self.default_formatters is None):
                    self.default_formatters = {}

        for param in params:
            if param in skipparams:
                continue

            outparams.append(param)

            colbase=param
            if param in parammapping:
                colbase=parammapping[param]

            outcol='{}{}{}'.format(prefix,colbase,suffix)
            outcols.append(outcol)

            if addformat2defaultformatter:
                if not(re.search('^N',param) is None) or param =='i':
                    self.default_formatters[outcol]='{:d}'.format
                elif param=='X2norm':
                    self.default_formatters[outcol]='{:.2f}'.format
                elif param in ['converged']:
                    pass
                else:
                    self.default_formatters[outcol]=format4outvals.format
            if setcol2None:
                if param in ['mean','mean_err','stdev','stdev_err','X2norm']:
                    self.t[outcol]=np.nan
                else:
                    self.t[outcol]=None

        return(set(zip(outparams,outcols)))

    def statresults2table(self,statparams,param2columnmapping,destindex=None):
        resultdict={}
        #statparams = copy.deepcopy(statparams)
        # loop through keys,outcols, and assign values
        for (param,outcol) in param2columnmapping:
            if destindex is None:
                resultdict[outcol]:statparams[param]
            else:
                self.t.loc[destindex,outcol]=statparams[param]

        # either put the data into a new row or into an existing one
        if destindex is None:
            outindex = self.t.newrow(resultdict)
        else:
            outindex = destindex
            #self.t.loc[destindex,outcols]=vals
        return(outindex)




#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Wed July 28 11:01:23 2021

@author: arest, bhilbert
"""
from astroquery.mast import Mast, Observations
from astropy.time import Time
from copy import deepcopy
import numpy as np
import pandas as pd
#import astropy.io.fits as fits
#from astropy.nddata import bitmask
import astroquery
import argparse,os,sys,re,copy,shutil
import yaml
from PIL import Image

# pdastroclass is wrapper around pandas.
from jwst_mast_query.pdastro import pdastroclass,unique,AnotB,AorB,AandB,split_commonpath

from jwst_mast_query.utils.constants import MAST_OBS_MODES, PARAM_DEFAULTS

# MAST API documentation:
# https://mast.stsci.edu/api/v0/pyex.html

# MAST CAOM Field Descriptions:
# https://mast.stsci.edu/api/v0/_c_a_o_mfields.html

# File naming schemes:
# https://jwst-pipeline.readthedocs.io/en/latest/jwst/data_products/file_naming.html
# https://jira.stsci.edu/browse/JSDP-1778


if astroquery.__version__<'0.4.2':
    raise RuntimeError("astroquery version=%s, at least 0.4.2 required!" % astroquery.__version__)

# environment variables:

# $MAST_API_TOKEN: needed to access the MAST database. If the token is not set
# with $MAST_API_TOKEN, then it needs to be specified with --token.
# it looks like that for publically available data (like the LRE data), no token is needed...

# $JWST_QUERY_CFGFILE: the config file is optional. It an be specified with
# $JWST_QUERY_CFGFILE or with --configfile

# List of allowed sca (detector) names
ALL_SCAS = ['a1', 'a2', 'a3', 'a4', 'a5', 'along',
            'b1', 'b2', 'b3', 'b4', 'b5', 'blong',
            'guider1','guider2',
            'nrs1','nrs2',
            'mirimage', 'mirifulong', 'mirifushort',
            'nis']


def getlimits(lims):
    if lims is None or len(lims)==0:
        return(None)
    if len(lims)==1:
        if re.search('\+$',lims[0]):
            lims[0]=re.sub('\+$','',lims[0])
            return([lims[0],None])
        elif  re.search('\-$',lims[0]):
            lims[0]=re.sub('\-$','',lims[0])
            return([None,lims[0]])
        else:
            return([lims[0],lims[0]])
    elif len(lims)==2:
        return([lims[0],lims[1]])
    else:
        raise RuntimeError(f'limits can have only 2 entries, more given! {lims}')

def addlink2string(s,link):
    return('<a href="%s">%s</a>' % (link,s))

def imagestring4web(imagename, width=None, height=None):
    imstring = '<img src="%s"' % imagename
    if height != None:
        imstring += f' height={height}'
        #if isinstance(height,int): height = str(height)
        #imstring += 'height=%s' % height
    if width != None:
        imstring += f' width={width}'
        #if isinstance(width,int): width = str(width)
        #imstring += 'width=%s' % width
    imstring +='>'
    return(imstring)

def image4table(imagename, TNimagename=None, width=None, height=None):
    if TNimagename is None:
        TNimagename=imagename
    return(addlink2string(imagestring4web(TNimagename,width=width,height=height),imagename))


def rmfile(filename,raiseError=1,gzip=False):
    " if file exists, remove it "
    if os.path.lexists(filename):
        os.remove(filename)
        if os.path.isfile(filename):
            if raiseError == 1:
                raise RuntimeError('ERROR: Cannot remove %s' % filename)
            else:
                return(1)
    if gzip and os.path.lexists(filename+'.gz'):
        os.remove(filename+'.gz')
        if os.path.isfile(filename+'.gz'):
            if raiseError == 1:
                raise RuntimeError('ERROR: Cannot remove %s' % filename+'.gz')
            else:
                return(2)
    return(0)


class query_mast:
    def __init__(self):
        '''
        Set server and web service endpoints for astroquery
        '''
        self.JwstObs = Mast()
        server = 'https://mast.stsci.edu'
        self.JwstObs._portal_api_connection.MAST_REQUEST_URL = server + "/portal_jwst/Mashup/Mashup.asmx/invoke"
        self.JwstObs._portal_api_connection.MAST_DOWNLOAD_URL = server + "/jwst/api/v0.1/download/file"
        self.JwstObs._portal_api_connection.COLUMNS_CONFIG_URL = server + "/portal_jwst/Mashup/Mashup.asmx/columnsconfig"
        self.JwstObs._portal_api_connection.MAST_BUNDLE_URL = server + "/jwst/api/v0.1/download/bundle"


        #self.RET_COLUMNS = ['proposal_id','dataURL','obsid','obs_id','t_min','t_exptime']
        self.SERVICES = {
                'SI_search': 'Mast.Jwst.Filtered.',
                'Caom_search':'Mast.Caom.Filtered',
                'Product_search':'Mast.Caom.Products'
                }

        self.verbose = 0
        self.debug = 0

        # delete all tables and outdir, set the default parameters
        self.reset()

    def reset(self):
        """
        delete all tables and outdir, set the default parameters into self.params

        Returns
        -------
        None.

        """
        self.outrootdir = None

        # observation and product table
        self.obsTable = pdastroclass()
        self.productTable = pdastroclass()
        # summary table: statistics/info for each propID/obsnum pair
        self.summary = pdastroclass()

        # indices of the selected products
        # They are sorted by self.params['sortcols_productTable'] (can also be set in config file)
        self.ix_selected_products = None

        # indices for obsTable, sorted by self.params['sortcols_obsTable'] (can also be set in config file)
        self.ix_obs_sorted = None
        # indices for summary table, sorted by self.params['sortcols_summaryTable'] (can also be set in config file)
        self.ix_summary_sorted = None

        # self.params will be populated from define_options()
        self.params = deepcopy(PARAM_DEFAULTS)

    def define_options(self, parser=None, usage=None, conflict_handler='resolve'):
        if parser is None:
            parser = argparse.ArgumentParser(usage=usage, conflict_handler=conflict_handler)

        parser.add_argument('-i', '--instrument', type=str, default=PARAM_DEFAULTS['instrument'], choices=['niriss','nircam','nirspec','miri','fgs'],
                            help='Specify the instrument (nircam, nirspec, niriss, miri, fgs)  (default=%(default)s)')
        parser.add_argument('--obsmode', type=str, nargs='+', default=PARAM_DEFAULTS['obsmode'], help=('Optionally specify the observation modes to query for. '
                                                                                                       'If provided, MAST will be queried only for these types of '
                                                                                                       'observations. e.g. assuming instrument is "nircam", '
                                                                                                       'obsmode: ["image", "wfss"] will query MAST for NIRCAM/IMAGE '
                                                                                                       'and NIRCAM/WFSS data. If left empty, the query will include '
                                                                                                       'all modes. For a list of valid modes, see the example config '
                                                                                                       'file in the repository. Note that as of 24 March 2023, MAST '
                                                                                                       'is in the process of reprocessing all data to add the mode '
                                                                                                       'values to the instrument names. Until this process is '
                                                                                                       'complete, some older data may not include the mode name, '
                                                                                                       'and therefore will not be found via a query that includes '
                                                                                                       'the mode name. (default=%(default)s)'))
        parser.add_argument('-v','--verbose', default=PARAM_DEFAULTS['verbose'], action='count')
        parser.add_argument('--propID', type=int, nargs="+", default=PARAM_DEFAULTS['propID'], help=('Specify the JWST proposal number to query for. This can be '
                                                                                                     'a 5 digit integer, or for a smaller number, an integer '
                                                                                                     'with or without prepended zeros. e.g. 1409 or 01409.'))
        parser.add_argument('--obsnums', type=int, nargs='+', default=PARAM_DEFAULTS['obsnums'], help=('Specify the observation numbers within the propID. This '
                                                                                                       'can be a single number, or a bracketed list of numbers. '
                                                                                                       'e.g. 3 or [3, 103] (default=%(default)s)'))
        parser.add_argument('--guidestars', action='store_true', default=PARAM_DEFAULTS['guidestars'], help=('If this is set to True, guidestar products are '
                                                                                                             'also included. Note: there are a **lot** of guide '
                                                                                                             'star products. We recommend you set to True only if '
                                                                                                             'really needed! (default=%(default)s)'))
        parser.add_argument('-f','--filetypes',  type=str, nargs="+", default=PARAM_DEFAULTS['filetypes'], help=('List of file types to select in the product table, '
                                                                                                                 'e.g., _uncal.fits or _uncal.jpg. If no suffix is '
                                                                                                                 'given, .fits is appended. If only letters, then _ '
                                                                                                                 'and .fits are added. For example, "uncal" gets '
                                                                                                                 'expanded to _uncal.fits. Typical image filetypes '
                                                                                                                 'are uncal, rate, rateints, cal. For downloading a '
                                                                                                                 'single file type, the brackets must still surround '
                                                                                                                 'the file suffix, as the script expects a list. A '
                                                                                                                 'relatively complete list of options includes: '
                                                                                                                 '["_segm.fits", "_asn.json", "_pool.csv", "_i2d.jpg", '
                                                                                                                 '"_thumb.jpg", "_cat.ecsv", "_i2d.fits", "_uncal.fits", '
                                                                                                                 '"_uncal.jpg", "_cal.fits", "_trapsfilled.fits", '
                                                                                                                 '"_cal.jpg", "_rate.jpg", "_rateints.jpg", '
                                                                                                                 '"_trapsfilled.jpg", "_rate.fits", "_rateints.fits"] '
                                                                                                                 'See the JWST calibration pipeline documentation for '
                                                                                                                 'a complete list. (default=%(default)s)'))
        parser.add_argument('--calib_levels',  type=int, nargs="+", default=PARAM_DEFAULTS['calib_levels'], help=('Only select products with the specified '
                                                                                                                   'calibration levels (calib_level column in '
                                                                                                                   'productTable) (default=%(default)s)'))
        parser.add_argument('--Nobs_per_batch', type=int, default=PARAM_DEFAULTS['Nobs_per_batch'], help=('For large programs, the query for the products can time '
                                                                                                          'out, and in these cases it is better to split up the query '
                                                                                                          'into Nobs_per_batch observations per batch. For example, '
                                                                                                          'if there are 22 observations, and Nobs_per_batch=4, then '
                                                                                                          'there will be 6 batches, 5 batches with 4 observations, '
                                                                                                          'and the last batch with 2. (default=%(default)s)'))
        parser.add_argument('-c','--configfile', type=str, default=PARAM_DEFAULTS['configfile'], help=('optional config file. default is set to $JWST_QUERY_CFGFILE. '
                                                                                                       'Use -vvv to see the full list of all set parameters.'))
        parser.add_argument('--login', default=PARAM_DEFAULTS['login'], nargs=2, help='username and password for login')
        parser.add_argument('--token', type=str, default=PARAM_DEFAULTS['token'], help=('MAST API token. You can also set the token in the environment '
                                                                                        'variable \$MAST_API_TOKEN'))
        parser.add_argument('--outrootdir', default=PARAM_DEFAULTS['outrootdir'], help=('The base directory for the downloaded products. This can be the name of '
                                                                                        'an environment variable (such as JWSTDOWNLOAD_OUTDIR), or '
                                                                                        'a path. Note the preceding "$" in the case where an environment variable is '
                                                                                        'given. If you include the --outrootdir command line argument when calling '
                                                                                        'jwst_query.py or jwst_download.py, that value will override any value '
                                                                                        'provided in the config file. (default=%(default)s)'))
        parser.add_argument('--outsubdir', default=PARAM_DEFAULTS['outsubdir'], help=('Any additional directory to add to the base directory. This can be used to '
                                                                                      'customize the organization of the downloaded products. (default=%(default)s)'))
        parser.add_argument('--obsnum2outsubdir', default=PARAM_DEFAULTS['obsnum2outsubdir'], help=('If True, the observation number will be used to create a '
                                                                                                    'subdirectory into which the appropriate files will be placed. '
                                                                                                    'e.g. $JWSTDOWNLOAD_OUTDIR/01410/obsnum23/'))
        parser.add_argument('--propIDs_obsnum2outsubdir', default=PARAM_DEFAULTS['propIDs_obsnum2outsubdir'], help=('Specify list of propID for which obsnum2outsubdir '
                                                                                                                    'is True. Only for the listed proposal numbers will '
                                                                                                                    'observation numbers be used to create a '
                                                                                                                    'subdirectory into which the appropriate files will '
                                                                                                                    'be placed. e.g. $JWSTDOWNLOAD_OUTDIR/01410/obsnum23/'))
        parser.add_argument('--skip_propID2outsubdir', action='store_true', default=PARAM_DEFAULTS['skip_propID2outsubdir'], help=('By default, the APT proposal ID '
                                                                                                                                   'is added as a subdir to the output '
                                                                                                                                   'directory. You can skip this with '
                                                                                                                                   'this option (default=%(default)s)'))
        parser.add_argument('--skip_check_if_outfile_exists', action='store_true', default=PARAM_DEFAULTS['skip_check_if_outfile_exists'],
                            help=('By default, the script queries MAST for products, and then checks if each file already exists in the output directory or not '
                                  '("dl_code" and "dl_str" columns). For large numbers of products, this can take time. By setting this option to True, this '
                                  'check can be skipped. (default=%(default)s)'))
        parser.add_argument('--skip_check_filesize', action='store_true', default=PARAM_DEFAULTS['skip_check_filesize'], help=('Don\'t check if output files have the '
                                                                                                                               'correct filesize. This makes it faster '
                                                                                                                               'for large lists, but files might be '
                                                                                                                               'corrupted.'))
        parser.add_argument('-e','--obsid_select', nargs="+", default=PARAM_DEFAULTS['obsid_select'], help=('Specify obsid range applied to "obsID" and "parent_obsid" '
                                                                                                            'columns in the PRODUCT table. If single value, then exact '
                                                                                                            'match. If single value has "+" or "-" at the end, then it '
                                                                                                            'is a lower and upper limit, respectively. If two values, '
                                                                                                            'then range. Examples: 385539+, 385539-, 385539 385600 '
                                                                                                            '(default=%(default)s)'))
        parser.add_argument('-l','--obsid_list', nargs="+", default=PARAM_DEFAULTS['obsid_list'], help=('Specify list of obsid applied to "obsID" and "parent_obsid" '
                                                                                                        'columns in the PRODUCT table. examples: 385539 385600 385530 '
                                                                                                        '(default=%(default)s)'))
        parser.add_argument('--sca', nargs="+", default=PARAM_DEFAULTS['sca'],
                            choices=['a1','a2','a3','a4','a5','along','b1','b2','b3','b4','b5','blong',
                                     'guider1','guider2','nrs1','nrs2','mirimage','mirifulong','mirifushort','nis'],
                                     help='Specify list of sca\'s to select. a5=along, b5=blong')
        parser.add_argument('--mastcolumns_obsTable', nargs="+", default=PARAM_DEFAULTS['mastcolumns_obsTable'], help=('Core columns returned from MAST to the obsTable. '
                                                                                                                       '(default=%(default)s)'))
        parser.add_argument('--outcolumns_obsTable', nargs="+", default=PARAM_DEFAULTS['outcolumns_obsTable'], help=('Output columns for the obsTable. '
                                                                                                                     '(default=%(default)s)'))
        parser.add_argument('--sortcols_obsTable', nargs="+", default=PARAM_DEFAULTS['sortcols_obsTable'], help=('The obsTable is sorted based on these columns. '
                                                                                                                       'The defaults sort the table in the order the '
                                                                                                                       'observations were observed. (default=%(default)s)'))
        parser.add_argument('--outcolumns_productTable', nargs="+", default=PARAM_DEFAULTS['outcolumns_productTable'], help=('List of columns to be shown in product '
                                                                                                                             'table (default=%(default)s)'))
        parser.add_argument('--sortcols_productTable', nargs="+", default=PARAM_DEFAULTS['sortcols_productTable'], help=('The productTable is sorted based on these '
                                                                                                                         'columns. The default sorts the table based '
                                                                                                                         'on calibration level. (default=%(default)s)'))
        parser.add_argument('--sortcols_summaryTable', nargs="+", default=PARAM_DEFAULTS['sortcols_summaryTable'], help=('The summary table is sorted based on these '
                                                                                                                         'columns. The default sorts the table in the '
                                                                                                                         'order the observations were observed. '
                                                                                                                         '(default=%(default)s)'))
        parser.add_argument('--jpg_separate_subdir', default=PARAM_DEFAULTS['jpg_separate_subdir'], help=('If True, downloaded jpgs are saved in separate "jpg" '
                                                                                                          'subdirectories, along side the fits files. (default=%(default)s)'))

        time_group = parser.add_argument_group("Time constraints for the observation/product search")
        time_group.add_argument('-l', '--lookbacktime', type=float, default=PARAM_DEFAULTS['lookbacktime'], help=('Lookback time in days. The script will query MAST '
                                                                                                                  'over a time from the lookback time to the present '
                                                                                                                  'moment. Note that all other time parameters '
                                                                                                                  '(date_select, etc) override the lookback time. '
                                                                                                                  '(default=%(default)s)'))
        time_group.add_argument('-d','--date_select', nargs="+", default=PARAM_DEFAULTS['date_select'], help=('Specify date range (MJD or isot format) applied to '
                                                                                                              '"dateobs_center" column. If single value, then exact '
                                                                                                              'match. If single value has "+" or "-" at the end, then '
                                                                                                              'it is a lower and upper limit, respectively. date_select '
                                                                                                              'will override the lookbacktime. Examples: '
                                                                                                              '58400+, 58400-,2020-11-23+, 2020-11-23 2020-11-25  '
                                                                                                              '(default=%(default)s)'))
        parser.add_argument('-s', '--savetables', type=str, default=PARAM_DEFAULTS['savetables'], help=('Save the tables (selected products, obsTable, summary with '
                                                                                                        'suffix selprod.txt, obs.txt, summary.txt, respectively) with '
                                                                                                        'the specified string as basename. Tables are saved in the '
                                                                                                        'same output directory as the data.  If no string is '
                                                                                                        'provided, the tables are not saved. (default=%(default)s)'))
        parser.add_argument('-w', '--makewebpages', action='store_true', default=PARAM_DEFAULTS['makewebpages'],
                            help=('Make webpages for the products for each propID containing info and images of the retrieved data. Note that this option is not '
                                  'currently in the config file, and must be specified at the command line. (default=%(default)s)'))
        parser.add_argument('--webpage_tablefigsize_width', default=PARAM_DEFAULTS['webpage_tablefigsize_width'], help=("Width in pixels of the resized jpg images "
                                                                                                                        "to be inserted into the index.html summary file."
                                                                                                                        "Recommended: 100-150, or don't specify if "
                                                                                                                        "webpage_mkthumbnails, since then the size of "
                                                                                                                        "the thumbnails are used by default. "
                                                                                                                        "(default=%(default)s)"))
        parser.add_argument('--webpage_tablefigsize_height', default=PARAM_DEFAULTS['webpage_tablefigsize_height'], help=("Height in pixels of the resized jpg images "
                                                                                                                          "to be inserted into the index.html summary "
                                                                                                                          "file. If left undefined, the height will be "
                                                                                                                          "determined from webpage_thumbnail_width and "
                                                                                                                          "the aspect ratio of the original image.  "
                                                                                                                          "(default=%(default)s)"))
        parser.add_argument('--webpage_level12_jpgs', nargs="+", default=PARAM_DEFAULTS['webpage_level12_jpgs'], help=('List of filetypes whose thumbnails will be shown in '
                                                                                                            'index.html. (default=%(default)s)'))
        parser.add_argument('--webpage_fitskeys2table', nargs="+", default=PARAM_DEFAULTS['webpage_fitskeys2table'], help=('List of fits header keywords that should '
                                                                                                                           'be copied to the table. (default=%(default)s)') )
        parser.add_argument('--webpage_cols4table', nargs="+", default=PARAM_DEFAULTS['webpage_cols4table'], help=('Columns to be shown in index.html. (default=%(default)s)'))
        parser.add_argument('--webpage_sortcols', nargs="+", default=PARAM_DEFAULTS['webpage_sortcols'], help=('Columns to sort index.html by. (default=%(default)s)'))
        parser.add_argument('--webpage_mkthumbnails', default=PARAM_DEFAULTS['webpage_mkthumbnails'], help=('If True, a thumbnail jpg is created for each of the '
                                                                                                            'jpg products listed in webpage_level12_jpgs. '
                                                                                                            '(default=%(default)s)'))
        parser.add_argument('--webpage_thumbnails_overwrite', default=PARAM_DEFAULTS['webpage_thumbnails_overwrite'], help=('If True, remake thumbnails even if they '
                                                                                                                            'already exist (default=%(default)s)'))
        parser.add_argument('--webpage_thumbnails_width', default=PARAM_DEFAULTS['webpage_thumbnails_width'], help=('Width in pixels of the resized jpg images '
                                                                                                                    'to be inserted into the index.html summary '
                                                                                                                    'file. (default=%(default)s)'))
        parser.add_argument('--webpage_thumbnails_height', default=PARAM_DEFAULTS['webpage_thumbnails_height'], help=('Height in pixels of the resized jpg images '
                                                                                                                      'to be inserted into the index.html summary '
                                                                                                                      'file. (default=%(default)s)'))
        return(parser)

    def get_arguments(self, args, configfile=None):
        '''

        Parameters
        ----------
        args : list
            pass the command line arguments to self.params.
            makes sure that the propID has the correct format (5 digit string)
        configfile : string, optional
            Config filename. The default is None. If None, then
            $JWST_QUERY_CONFIGFILE is used if exists.

        Returns
        -------
        None.

        '''

        def subenvvarplaceholder(paramsdict):
            """ Loop through all string parameters and substitute environment variables. environment variables have the form $XYZ """
            envvarpattern=re.compile('\$(\w+)')

            for param in paramsdict:
                if isinstance(paramsdict[param], str):
                    envvarnames=envvarpattern.findall(paramsdict[param])
                    if envvarnames:
                        for name in envvarnames:
                            if not (name in os.environ):
                                if param=='outrootdir':
                                    # this is a special hack: if outrootdir is not defined, it is set to '.'
                                    print(f'WARNING! environment variable JWSTDOWNLOAD_OUTDIR in {param} not defined! Thus setting {param} to None...')
                                    paramsdict[param] = None
                                else:
                                    raise RuntimeError("environment variable %s used in config file, but not set!" % name)
                            else:
                                envval=os.environ[name]
                                subpattern='\$%s' % (name)
                                paramsdict[param] = re.sub(subpattern,envval,paramsdict[param])
                elif isinstance(paramsdict[param], dict):
                #elif type(dict[param]) is types.DictType:
                    # recursive: sub environment variables down the dictionary
                    subenvvarplaceholder(paramsdict[param])
            return(0)

        # get the parameters from the config file
        if args.configfile is not None:
            if not os.path.isfile(args.configfile):
                raise RuntimeError('config file %s does not exist!' % (args.configfile))
            print(f'Loading config file {args.configfile}')
            cfgparams = yaml.load(open(args.configfile,'r'), Loader=yaml.FullLoader)
            self.params.update(cfgparams)
            self.params['configfile'] = args.configfile

            if args.verbose>2:
                print('\n### CONFIG FILE PARAMETERS:')
                for p in cfgparams:
                    print('config file args: setting %s to' % (p),cfgparams[p])

            # Make sure propID and obsnums, if they have been entered in the config file,
            # are lists, rather than integers. When entered via command line, they are
            # guaranteed to be lists.
            list_keys = ['propID', 'obsnums']
            for key in list_keys:
                if key in self.params.keys():
                    if not isinstance(self.params[key], list) and self.params[key] is not None:
                        self.params[key] = [self.params[key]]

            # Ensure that the propID is a 5-digit string
            if 'propID' in self.params.keys():
                if self.params['propID'] is not None:
                    self.params['propID'] = ['%05d' % (int(self.params['propID'][0]))]

        subenvvarplaceholder(self.params)

        # Make sure obsmode is a list, even if it's just [None]
        list_keys = ['obsmode']
        for key in list_keys:
            if key in self.params.keys():
                if not isinstance(self.params[key], list):
                    self.params[key] = [self.params[key]]

        # Go through optional parameters.
        # 'None' does not overwrite previously set parameters (either default or config file)
        if args.verbose>2:
            print('\n### OPTIONAL COMMAND LINE PARAMETERS:')
        argsdict = vars(args)
        for arg in argsdict:

            # skip config file
            if arg=='configfile': continue

            if argsdict[arg] != PARAM_DEFAULTS[arg]:
                if args.verbose>2:
                    print('optional args: setting %s to %s' % (arg,argsdict[arg]))
                self.params[arg]=argsdict[arg]
            else:
                if arg not in  self.params:
                    self.params[arg]=None

        if 'obsnums' not in self.params.keys():
            self.params['obsnums']=None

        # propID
        if self.params['propID'] is not None:
            if len(self.params['propID'])>1:
                self.params['obsnums'] = self.params['propID'][1:]
            self.params['propID'] = '%05d' % (int(self.params['propID'][0]))
        self.verbose = self.params['verbose']

        if self.verbose>2:
            print('\n### FINAL PARAMETERS:')
            for p in self.params:
                print(f'{p}: {self.params[p]}')

        print('INSTRUMENT: ', self.params['instrument'])
        print('obsmode: ', self.params['obsmode'])
        print('propID: ',self.params['propID'])
        print('obsnums: ',self.params['obsnums'])

        return(0)

    def login(self,token=None,raiseErrorFlag=False):
        if token is None:
            token = self.params['token']

        if token is None:
            if raiseErrorFlag:
                raise RuntimeError("No token!! Cannot login! you can set the token with \$API_MAST_TOKEN, as 'token' in the config file, or with --token ")
            else:
                print("Warning: no token given, could not login, only public requests possible")
                return(1)
        else:
            self.JwstObs.login(token=token, store_token=False)
        return(0)

    def obsid_select(self,obsid_select,
                     productTable=None,
                     ix_selected_products=None):

        if productTable is None:
           productTable=self.productTable

        if ix_selected_products is None:
            ix_selected_products = self.ix_selected_products

        # parse trailing '+' and '-', and get limits
        limits = getlimits(obsid_select)
        if limits is None:
            return(ix_selected_products)

        for i in range(len(limits)):
            if limits[i] is not None: limits[i]=int(limits[i])

        ixs = self.productTable.ix_not_null('obsID',indices=ix_selected_products)
        ixs_keep1 = self.productTable.ix_inrange('obsID',limits[0],limits[1],indices=ixs)
        ixs_keep2 = self.productTable.ix_inrange('parent_obsid',limits[0],limits[1],indices=ixs)
        ixs_keep = unique(AorB(ixs_keep1,ixs_keep2))
        print(f'obsid cut {limits[0]} - {limits[1]}: keeping {len(ixs_keep)} from {len(ixs)}')
        return(ixs_keep)

    def obsid_list(self,obsid_list,
                     productTable=None,
                     ix_selected_products=None):

        if productTable is None:
           productTable=self.productTable

        if ix_selected_products is None:
            ix_selected_products = self.ix_selected_products

        if obsid_list is None or len(obsid_list)==0:
            return(ix_selected_products)

        ixs_keep = []
        for obsid in obsid_list:
            obsid = int(obsid)
            ixs2add = self.productTable.ix_equal('obsID',obsid,indices=ix_selected_products)
            ixs_keep.extend(ixs2add)
            ixs2add = self.productTable.ix_equal('parent_obsid',obsid,indices=ix_selected_products)
            ixs_keep.extend(ixs2add)
        ixs_keep = unique(ixs_keep)
        print(f'obsid list cut: keeping {len(ixs_keep)} from {len(ix_selected_products)}')
        return(ixs_keep)


    def select_sca(self, sca_list,
                   productTable=None,
                   ix_selected_products=None):

        if productTable is None:
           productTable=self.productTable

        if ix_selected_products is None:
            ix_selected_products = self.ix_selected_products

        if sca_list is None or len(sca_list)==0:
            return(ix_selected_products)

        if isinstance(sca_list,str):
            sca_list=[sca_list]

        ixs_keep = []
        for sca in sca_list:
            if sca=='a5': sca='along'
            if sca=='b5': sca='blong'
            ixs2add = self.productTable.ix_equal('sca',sca,indices=ix_selected_products)
            ixs_keep.extend(ixs2add)
        ixs_keep = unique(ixs_keep)
        print(f'select sca {sca_list}: keeping {len(ixs_keep)} from {len(ix_selected_products)}')
        return(ixs_keep)

    def select_obsnums(self, obsnum_list,
                       productTable=None,
                       ix_selected_products=None):

        if productTable is None:
           productTable=self.productTable

        if ix_selected_products is None:
            ix_selected_products = self.ix_selected_products

        if obsnum_list is None or len(obsnum_list)==0:
            return(ix_selected_products)

        if isinstance(obsnum_list,str):
            obsnum_list=[obsnum_list]

        ixs_keep = []
        for obsnum in obsnum_list:
            ixs2add = self.productTable.ix_equal('obsnum',obsnum,indices=ix_selected_products)
            ixs_keep.extend(ixs2add)
        ixs_keep = unique(ixs_keep)
        print(f'select obsnum {obsnum_list}: keeping {len(ixs_keep)} from {len(ix_selected_products)}')
        return(ixs_keep)


    def get_mjd_limits(self, lookbacktime=None, date_select=None):
        """Determine the earliest and latest dates to use for the query

        Parameters
        ----------
        lookbacktime : int
            Number of days before the current date to search

        date_select : str
            String of starting and stopping dates to use

        Returns
        -------
        mjd_min : float
            MJD of the earliest time to search
        mjd_max : float
            MJD of the latest time to search
        """
        if lookbacktime is None: lookbacktime = self.params['lookbacktime']

        if date_select is None: date_select = self.params['date_select']

        # If date_select is a string, split it into a list.
        # When provided at the command line, date_select is a list.
        # When provided in cfg file, date_select is a string.
        if isinstance(date_select, str):
            date_select = date_select.split(' ')

        mjd_min = mjd_max = None
        # parse trailing '+' and '-', and get limits
        limits = getlimits(date_select)
        if limits is not None:
            # Convert dates into MJD if necessary
            for i in (0,1):
                if limits[i] is not None:
                    try:
                        limits[i] = float(limits[i])
                    except:
                        limits[i]= Time(limits[i], format='isot').to_value('mjd')
            mjd_min = limits[0]
            mjd_max = limits[1]
        else:
            if (mjd_min is None):
                if self.params['lookbacktime']>0.0:
                    if self.verbose>1: print('Note: Looking back %.1f days' % (self.params['lookbacktime']))
                    mjd_min = Time.now().mjd - self.params['lookbacktime']

        if mjd_max is None:
            mjd_max = Time.now().mjd+0.1

        return(mjd_min, mjd_max)

    def observation_query(self, propID=None, instrument=None, mjd_min=None, mjd_max=None, token=None):
        '''
        Perform query for observations matching JWST instrument and program ID
        that began after a particular date.

        Parameters
        ----------
        propID : int
            Proposal ID number
        instrument : str
            Instrument name, e.g. 'nircam'
        mjd_min : float
            MJD of the earliest date to search
        mjd_max : float
            MJD of the latest date to search

        Returns
        -------
        self.obsTable
        '''

        if instrument is None:
            instrument = self.params['instrument']

        print('MJD range:',mjd_min, mjd_max)

        # Work-around for MAST update, where instrument name being changed to
        # a string of the format "<instrument>/<mode>" rather than "<instrument>"
        # e.g "MIRI/IMAGE" rather than "MIRI". Wildcards are not yet supported
        # by astroquery when working with the JwstFiltered data, so we need to
        # get a list of all possible modes associated with the proposal, and then
        # query for each separately

        # To get the list of modes, we can query using Observations, which does support
        # wildcard characters, but returnes only limited information. In this case, we
        # care only about the instrument_name values in the results.

        # Check to see if the user entered a list of observing modes. If not, we
        # query all modes for the instrument via wildcard
        if self.params['obsmode'][0] is None or len(self.params['obsmode']) == 0:
            print(f'No obsmode given. Querying for all files for {instrument}.')
            inst_list = list(set(Observations.query_criteria(instrument_name=f'{instrument}*',
                                                             t_min=[mjd_min, mjd_max])['instrument_name']))
        else:
            initial_inst_list = [f'{instrument.upper()}/{mode.upper()}' for mode in self.params['obsmode']]
            inst_list = deepcopy(initial_inst_list)

            # Check against the set of allowed modes, and alert the user to any invalid entries
            # The MAST search will run without crashing if there is a bad entry. It will simply
            # ignore the entry.
            valid_modes = [mode.split('/')[1].lower() for mode in MAST_OBS_MODES[instrument]]
            for mode in initial_inst_list:
                if mode not in MAST_OBS_MODES[instrument]:
                    mode_only = mode.split('/')[1].lower()
                    print(f'\n\nWARNING: "{mode_only}" is not a valid instrument mode for the MAST search.')
                    print('This entry will be ignored in the query.')
                    print(f'Valid {instrument} modes are: {valid_modes}\n\n')
                    inst_list.remove(mode)

            # If all of the user-input obsmodes are invalid, then we fall back to using the wildcard
            if len(inst_list) == 0:
                print('All entered modes are invalid. Falling back to query across all possible obsmodes.')
                inst_list = list(set(Observations.query_criteria(instrument_name=f'{instrument}*',
                                                             t_min=[mjd_min, mjd_max])['instrument_name']))

        # If inst_list is empty, then there are no observations to query for
        if len(inst_list) == 0:
            return(0)

        columns = ','.join(self.params['mastcolumns_obsTable'])
        service = self.SERVICES['Caom_search']
        params = {"columns":columns,
                  "filters":[
                  {"paramName":"obs_collection","values":["JWST"]},
                  {"paramName":"instrument_name","values":inst_list},
                  {"paramName":"t_min",
                    "values":[{"min":mjd_min,"max":mjd_max}]},
                  ]}

        # Only add propID entry if not None. If it is set to None it doesn't work!
        if propID is not None:
            params["filters"].append({"paramName":"proposal_id","values":[propID]})

        if self.verbose:
            print('\n#### Querying Obstable....')
        if self.verbose>1:
            print('query params:',params)

        if token is not None:
            self.obsTable.t = self.JwstObs.service_request(service, params, mast_token=token).to_pandas()
        else:
            self.obsTable.t = self.JwstObs.service_request(service, params).to_pandas()
        if self.verbose:
            print('Obstable obtained: length:',len(self.obsTable.t))

        if len(self.obsTable.t)==0:
            print('WARNING!! No observations found!!')
            return(0)

        # Find the obsnum # from the filename if possible.
        ixs = self.obsTable.getindices()
        obsnumsearch = re.compile('^jw\d{5}(\d{3})\d{3}\_')
        self.obsTable.t['obsnum']=pd.NA
        for ix in ixs:
            m = obsnumsearch.search(self.obsTable.t.loc[ix,'obs_id'])
            if m is not None:
                self.obsTable.t.loc[ix,'obsnum']=int(m.groups()[0])
            else:
                self.obsTable.t.loc[ix,'obsnum']=pd.NA

        self.obsTable.mjd2dateobs('t_min','date_min')

        self.ix_obs_sorted = self.obsTable.ix_sort_by_cols(self.params['sortcols_obsTable'])

        # make proposal_id integer
        if 'proposal_id' in self.obsTable.t.columns:
            self.obsTable.t['proposal_id']=self.obsTable.t['proposal_id'].astype('int')
        if 'obsid' in self.obsTable.t.columns:
            self.obsTable.t['obsid']=self.obsTable.t['obsid'].astype('int')

        self.obsTable.t.style.set_properties(subset=["obs_id"], **{'text-align': 'left'})
        #self.obsTable.default_formatters['obs_id']='{:<}'.format


        if self.verbose>1: print('Obstable columns:',self.obsTable.t.columns)

        return self.obsTable

    def fix_obsnum(self, obsTable=None, productTable=None):
        """
        fix obsnum=NA in productTable and obsTable

        Parameters
        ----------
        obsTable : pdastroclass, optional
            Table with observations. The default is None. If None, then self.obsTable is used
        productTable : pdastroclass, optional
            Table with products. The default is None. If None, then self.productTable is used
        Returns
        -------
        None.

        """
        if obsTable is None:
            obsTable=self.obsTable

        if productTable is None:
           productTable=self.productTable

        self.obsTable.t['obsnum']=pd.Series(self.obsTable.t['obsnum'],dtype=pd.Int32Dtype())

        ixs_obsTable = obsTable.getindices()
        for ix_obsTable in ixs_obsTable:
            # find all products for a given observation
            obsid = obsTable.t.loc[ix_obsTable,'obsid']
            if self.verbose>2:
                print('### obsid',obsid)
            ixs_prodTable = productTable.ix_equal('parent_obsid',obsid)
            if self.verbose>2:
                productTable.write(indices=ixs_prodTable)

            # If the obsnum cannot be determined from the obsTable 'obs_id', then get it from the products!
            if obsTable.t.loc[ix_obsTable,'obsnum'] is pd.NA:
                # remove None entries
                ixs = productTable.ix_not_null('obsnum',indices=ixs_prodTable)
                if len(ixs)>0:
                    obsnum = unique(productTable.t.loc[ixs,'obsnum'])
                    if len(obsnum)==0:
                        obsTable.write(indices=ixs_obsTable)
                        productTable.write(indices=ixs_prodTable)
                        raise RuntimeError('No obsnum for these observations and products!')
                    """
                    if len(obsnum)>1:
                        print('\n BUUUGGGGG!!!!')
                        obsTable.write(indices=[ix_obsTable])
                        productTable.write(indices=ixs)
                        print('More than one obsnum!',obsnum)
                        raise RuntimeError('BUGGG!!!!!')
                    """
                    if len(obsnum)==1:
                        obsTable.t.loc[ix_obsTable,'obsnum'] = obsnum
                    else:
                        if self.debug:
                            obsTable.write(indices=[ix_obsTable])
                            productTable.write(indices=ixs)
                            print('More than one obsnum!',obsnum)
                            raise RuntimeError('BUGGG!!!!!')

                else:
                    print('Warning: could not determine obsnum for observation %s' % (obsTable.t.loc[ix_obsTable,'obsid']))

            # If there are entries in productTable for which 'obsnum' is None (happens to some calib_level=3 products) fill them up
            ixs_null = productTable.ix_is_null('obsnum',indices=ixs_prodTable)
            if len(ixs_null)>0:
                productTable.t.loc[ixs_null,'obsnum'] = obsTable.t.loc[ix_obsTable,'obsnum']


    def product_query_bkp(self, obsTable=None, guidestars=None):
        '''
        Perform query for data products based on obs_id's in observation table
        '''

        if obsTable is None:
            obsTable=self.obsTable

        if guidestars is None:
            guidestars = self.params['guidestars']
            if guidestars is None:
                guidestars = False


        # query MAST for all products for the obsid's
        obsids = ','.join(obsTable.t['obsid'].astype('str'))
        service = self.SERVICES['Product_search']
        params = {"obsid":obsids,
                  "columns":['type','productType'],
                  "format":"json"
                  }
        if self.verbose:
            print('\n#### Querying ProductTable....')
        self.productTable.t = self.JwstObs.service_request(service, params).to_pandas()
        if self.verbose:
            print('productTable obtained: length:',len(self.productTable.t))
            #print('productTable columns:',self.productTable.t.columns)

        # remove guide stars if wanted...
        if not guidestars:
            ixs_products = self.productTable.getindices()
            gs_text = '_gs-'
            ixs_gs = self.productTable.ix_matchregex('productFilename',gs_text)
            ixs_keep_products = AnotB(ixs_products,ixs_gs)
            self.productTable.t=self.productTable.t.loc[ixs_keep_products].reset_index()
            if self.verbose:
                print('Removing %d guide star products from a total of %d products, %d left' % (len(ixs_gs),len(ixs_products),len(ixs_keep_products)))

        # Fill the suffix column with the suffix of the form _bla1.bla2, e.g. _uncal.fits
        # This will later be used to figure out
        self.productTable.t['filetype'] = self.productTable.t['productFilename'].str.extract(r'(\_[a-zA-Z0-9]+\.[a-zA-Z0-9]+)$')

        scas = list(np.repeat(np.nan, len(self.productTable.t['obs_id'])))
        for i, oid in enumerate(self.productTable.t['obs_id']):
            match = [word for word in ALL_SCAS if word in oid]
            if len(match) > 0:
                scas[i] = match[0]
        self.productTable.t['sca'] = scas

        # Find the obsnum # from the filename if possible.
        ixs = self.productTable.getindices()
        obsnumsearch = re.compile('^jw\d{5}(\d{3})\d{3}\_')
        self.productTable.t['obsnum']=pd.NA
        for ix in ixs:
            m = obsnumsearch.search(self.productTable.t.loc[ix,'obs_id'])
            if m is not None:
                self.productTable.t.loc[ix,'obsnum']=int(m.groups()[0])
            else:
                self.productTable.t.loc[ix,'obsnum']=pd.NA

        # make proposal_id integer
        self.productTable.t['proposal_id']=self.productTable.t['proposal_id'].astype('int')
        # make obsid integer
        if 'obsID' in self.productTable.t.columns:
            self.productTable.t['obsID']=self.productTable.t['obsID'].astype('int')
        if 'parent_obsid' in self.productTable.t.columns:
            self.productTable.t['parent_obsid']=self.productTable.t['parent_obsid'].astype('int')

        self.fix_obsnum(obsTable=obsTable, productTable=self.productTable)

        if self.verbose: print('productTable columns:',self.productTable.t.columns)
        if self.verbose:
            allfiletypes = unique(self.productTable.t['filetype'])
            print('List of all filetypes of obtained products:',allfiletypes)

        return self.productTable

    def product_query(self, obsTable=None, guidestars=None, Nobs_per_batch=2):
        '''
        Perform query for data products based on obs_id's in observation table
        '''

        if obsTable is None:
            obsTable=self.obsTable

        if guidestars is None:
            guidestars = self.params['guidestars']
            if guidestars is None:
                guidestars = False


        # query MAST for all products for the obsid's
        obsids = ','.join(obsTable.t['obsid'].astype('str'))
        service = self.SERVICES['Product_search']
        #params = {"obsid":obsids,
        #          "columns":['type','productType'],
        #          "format":"json"
        #          }
        if Nobs_per_batch is not None:
            if self.verbose:
                print(f'\n#### Querying ProductTable in Nobs={Nobs_per_batch} per batch ....')

            obsid_list = list(obsTable.t['obsid'].astype('str'))
            obsids_chunks = [obsid_list[i:i+Nobs_per_batch] for i in range(0, len(obsid_list), Nobs_per_batch)]
            tmptables = []
            if self.verbose:
                print(f'{len(obsids_chunks)} batches')
            for obsids_chunk in obsids_chunks:
                obsids = ','.join(obsids_chunk)
                if self.verbose:
                    print(f'Querying products for obsids={obsids}')
                params = {"obsid":obsids,
                          "columns":['type','productType'],
                          "format":"json"
                          }
                tmptable = self.JwstObs.service_request(service, params).to_pandas()

                # Remove duplicate rows based on the productFilename value
                tmptable.drop_duplicates(subset='productFilename', keep='first', inplace=True, ignore_index=True)

                if self.verbose:
                    print(f'After filtering duplicate entries, {len(tmptable)} products found for this chunk of observations')
                tmptables.append(tmptable)

            self.productTable.t = pd.concat(tmptables, axis=0, ignore_index=True)

            # Remove duplicate rows that come from different queries
            self.productTable.t.drop_duplicates(subset='productFilename', keep='first', inplace=True, ignore_index=True)
        else:
            params = {"obsid":obsids,
                      "columns":['type','productType'],
                      "format":"json"
                      }
            if self.verbose:
                print('\n#### Querying ProductTable....')
            self.productTable.t = self.JwstObs.service_request(service, params).to_pandas()
        #self.productTable.write()
        if self.verbose:
            print(f'productTable obtained with {len(self.productTable.t)} entries')

        # remove guide stars if wanted...
        if not guidestars:
            ixs_products = self.productTable.getindices()
            gs_text = '_gs-'
            ixs_gs = self.productTable.ix_matchregex('productFilename',gs_text)
            ixs_keep_products = AnotB(ixs_products,ixs_gs)
            self.productTable.t=self.productTable.t.loc[ixs_keep_products].reset_index()
            if self.verbose:
                print('Removing %d guide star products from a total of %d products, %d left' % (len(ixs_gs),len(ixs_products),len(ixs_keep_products)))

        # Fill the suffix column with the suffix of the form _bla1.bla2, e.g. _uncal.fits
        # This will later be used to figure out
        self.productTable.t['filetype'] = self.productTable.t['productFilename'].str.extract(r'(\_[a-zA-Z0-9]+\.[a-zA-Z0-9]+)$')

        scas = list(np.repeat(np.nan, len(self.productTable.t['obs_id'])))
        for i, oid in enumerate(self.productTable.t['obs_id']):
            match = [word for word in ALL_SCAS if word in oid]
            if len(match) > 0:
                scas[i] = match[0]
        self.productTable.t['sca'] = scas

        #bla = self.productTable.t['obs_id'].str.extract(r'^jw\d{8}(\d{3})_(\d{2})')
        #print(bla,bla.columns,bla[0])
        # get the visit number
        self.productTable.t['visit']= self.productTable.t['obs_id'].str.extract(r'^jw\d{8}(\d{3})_\d{2}\d{1}\d{2}_')
        #self.productTable.t['vstgrp']= self.productTable.t['obs_id'].str.extract(r'^jw\d{8}\d{3}_(\d{2})\d{1}\d{2}_')

        # mark all products that are fits or jpg in separate columns
        self.productTable.t['fits']=self.productTable.t['jpg']=False
        ixs_fits = self.productTable.ix_matchregex('filetype',r'\.fits$')
        self.productTable.t.loc[ixs_fits,'fits']=True
        ixs_fits = self.productTable.ix_matchregex('filetype',r'\.jpg$')
        self.productTable.t.loc[ixs_fits,'jpg']=True

        # Find the obsnum # from the filename if possible.
        ixs = self.productTable.getindices()
        obsnumsearch = re.compile('^jw\d{5}(\d{3})\d{3}\_')
        self.productTable.t['obsnum']=pd.NA
        for ix in ixs:
            m = obsnumsearch.search(self.productTable.t.loc[ix,'obs_id'])
            if m is not None:
                self.productTable.t.loc[ix,'obsnum']=int(m.groups()[0])
            else:
                self.productTable.t.loc[ix,'obsnum']=pd.NA

        # make proposal_id integer
        self.productTable.t['proposal_id']=self.productTable.t['proposal_id'].astype('int')
        # make obsid integer
        if 'obsID' in self.productTable.t.columns:
            self.productTable.t['obsID']=self.productTable.t['obsID'].astype('int')
        if 'parent_obsid' in self.productTable.t.columns:
            self.productTable.t['parent_obsid']=self.productTable.t['parent_obsid'].astype('int')

        self.fix_obsnum(obsTable=obsTable, productTable=self.productTable)

        if self.verbose: print('productTable columns:',self.productTable.t.columns)
        if self.verbose:
            allfiletypes = unique(self.productTable.t['filetype'])
            print('List of all filetypes of obtained products:',allfiletypes)

        return self.productTable


    def product_filter(self, productTable=None, filetypes=None, calib_levels=None, gs_omit=True):
        '''
        Filter the list of products based on the filetype (or better suffix)
        '''
        if productTable is None:
           productTable=self.productTable

        if filetypes is None:
            filetypes=self.params['filetypes']

        if calib_levels is None:
            calib_levels=self.params['calib_levels']

        if self.verbose:
            print('\n#### Selecting products....')

        # if necessary, add leading '_' and suffix '.fits'
        if filetypes is not None:
            for i in range(len(filetypes)):
                if filetypes[i] in ['fits','jpg']:
                    filetypes[i]+='$'
                    continue
                if re.search('^_',filetypes[i]) is None:
                    filetypes[i] = '_'+filetypes[i]
                if re.search('\.',filetypes[i]) is None:
                    filetypes[i] += '.fits'

            print('allowed filetype list:',filetypes)

        ix_products = self.productTable.getindices()
        # remove guide stars if wanted...
        if gs_omit:
            gs_text = '_gs-'
            ix_gs = self.productTable.ix_matchregex('productFilename',gs_text)
            if self.verbose>1:
                print('Removing %d guide star products from a total of %d products, %d left' % (len(ix_gs),len(ix_products),len(ix_products)-len(ix_gs)))
            ix_products = AnotB(ix_products,ix_gs)


        if calib_levels is not None:
            ix_calib_level = []
            for calib_level in calib_levels:
                ixs = self.productTable.ix_equal('calib_level',calib_level,indices=ix_products)
                if self.verbose>1:
                    print('Keeping %d products with calib_level %d' % (len(ixs),calib_level))
                ix_calib_level.extend(ixs)
            if self.verbose>1:
                print('Keeping %d out of %d products with the correct calib_levels' % (len(ix_calib_level),len(ix_products)))
            ix_products = ix_calib_level

        # Loop trough the filetypes and get all entries from ix_products list
        self.ix_selected_products = []
        if filetypes is not None:
            for filetype in filetypes:

                # make sure the '.' in the regular expression is literal, and also add '$' to the end
                regex=re.sub('\.','\.',filetype)+'$'

                # get the indices for matching filetype ...
                ix_matching_filetype = self.productTable.ix_matchregex('filetype',regex,indices=ix_products)
                # ... and add them to the list of good indices
                self.ix_selected_products.extend(ix_matching_filetype)

                if self.verbose>1:
                    print('%d products with filetype regex matching %s' % (len(ix_matching_filetype),regex))
            if self.verbose:
                print('%d products with correct filetypes left' % (len(self.ix_selected_products)))
        else:
            self.ix_selected_products = ix_products

        # get the list of valid filetypes that exist
        self.params['filetypes'] = unique(self.productTable.t.loc[self.ix_selected_products,'filetype'])

        self.ix_selected_products = self.productTable.ix_sort_by_cols(self.params['sortcols_productTable'],indices=self.ix_selected_products)


        return(self.ix_selected_products)

    def update_obstable_indices(self, ixs_prod, ix_obs_sorted=None, obsTable=None, productTable=None):
        if obsTable is None:
            obsTable=self.obsTable

        if productTable is None:
           productTable=self.productTable

        if ix_obs_sorted is None:
            ix_obs_sorted = self.ix_obs_sorted

        # get all parent_obsids of the selected products
        obsids = unique(productTable.t.loc[ixs_prod,'parent_obsid'])

        # get all keys in observation table that are associated with these obsids
        new_ix_obs_sorted = []
        for obsid in obsids:
            new_ix_obs_sorted.extend(self.obsTable.ix_equal('obsid',obsid,indices=ix_obs_sorted))

        # Keep the order from ix_obs_sorted!
        new_ix_obs_sorted = AandB(ix_obs_sorted,new_ix_obs_sorted)

        return(new_ix_obs_sorted)

    def get_outdir(self,productTable, ix,
                   outdir=None,
                   skip_propID2outsubdir=False,
                   obsnum2outsubdir=False,
                   jpg_separate_subdir=False):

        if outdir is None:
            outdir = self.outrootdir
        if outdir is None:
            raise RuntimeError("outdir is not defined!")
        outdir = outdir.rstrip("/")


        if not skip_propID2outsubdir:
            outdir += f'/{productTable.t.loc[ix,"proposal_id"]:05d}'
            #outdir += '/{:05d}'.format(productTable.t.loc[ix,"proposal_id"])

        # shameless hack: for NIRCam
        if obsnum2outsubdir or (int(productTable.t.loc[ix,"proposal_id"]) in self.params["propIDs_obsnum2outsubdir"]):
            if productTable.t.loc[ix,"obsnum"] is pd.NA:
                outdir += '/NA'
            else:
                outdir += f'/obsnum{productTable.t.loc[ix,"obsnum"]:02d}'

        if jpg_separate_subdir and productTable.t.loc[ix,"jpg"]:
            outdir += '/jpg'

        return(outdir)


    def mk_outfilename(self, productTable, ix,
                       outdir=None,
                       skip_propID2outsubdir=False,
                       obsnum2outsubdir=False,
                       jpg_separate_subdir=False,
                       skip_check_if_outfile_exists=False,
                       skip_check_filesize=False):

        # get the output directory
        outdir = self.get_outdir(productTable, ix, outdir=outdir,
                                 skip_propID2outsubdir=skip_propID2outsubdir,
                                 obsnum2outsubdir=obsnum2outsubdir,
                                 jpg_separate_subdir=jpg_separate_subdir)


        #if outdir is None:
        #    outdir = self.outrootdir
        #if outdir is None:
        #    raise RuntimeError("outdir is not defined!")
        #outdir.rstrip("/")


        #if not skip_propID2outsubdir:
        #    outdir += '/{:05d}'.format(productTable.t.loc[ix,"proposal_id"])

        outfilename = productTable.t.loc[ix,'productFilename']

        fullfilename = os.path.abspath(f'{outdir}/{outfilename}')
        productTable.t.loc[ix,'outfilename']=fullfilename

        if not skip_check_if_outfile_exists:
            if os.path.exists(fullfilename):
                if not skip_check_filesize:
#                    print('VVV',productTable.t.loc[ix,'outfilename'],productTable.t.loc[ix,'size'],os.path.getsize(productTable.t.loc[ix,'outfilename']))
                    outfile_filesize = os.path.getsize(productTable.t.loc[ix,'outfilename'])
                    if outfile_filesize != int(productTable.t.loc[ix,'size']):
                        productTable.t.loc[ix,'dl_code'] = 4
                        productTable.t.loc[ix,'dl_str'] = 'ERRfilesize'
                    else:
                        productTable.t.loc[ix,'dl_code'] = 2
                        productTable.t.loc[ix,'dl_str'] = 'exists'
                else:
                    productTable.t.loc[ix,'dl_code'] = 2
                    productTable.t.loc[ix,'dl_str'] = 'exists'

            else:
                productTable.t.loc[ix,'dl_code'] = 0
                productTable.t.loc[ix,'dl_str'] = None


    def set_outrootdir(self,outrootdir=None,outsubdir=None):
        self.outrootdir = outrootdir
        if self.outrootdir is None:
            self.outrootdir = self.params['outrootdir']
        # if outdir is not defined, use '.'
        if self.outrootdir is None or self.outrootdir == '':
            self.outrootdir = '.'

        if outsubdir is not None and outsubdir!='':
            self.outrootdir  += f'/{outsubdir}'
        elif self.params['outsubdir'] is not None and self.params['outsubdir']!='':
            self.outrootdir  += f'/{self.params["outsubdir"]}'

        self.outrootdir = os.path.abspath(self.outrootdir)
        return(self.outrootdir)


    def mk_outfilenames(self,
                        productTable=None,
                        ix_selected_products=None,
                        outdir=None,
                        skip_propID2outsubdir=False,
                        obsnum2outsubdir=False,
                        jpg_separate_subdir=False,
                        skip_check_if_outfile_exists=False,
                        skip_check_filesize=False):

        if productTable is None:
           productTable=self.productTable

        if ix_selected_products is None:
            ix_selected_products = self.ix_selected_products



        productTable.t.loc[ix_selected_products,'outfilename'] = None
        productTable.t.loc[ix_selected_products,'dl_code'] = pd.NA
        productTable.t.loc[ix_selected_products,'dl_str'] = None
        productTable.t['dl_code'] = productTable.t['dl_code'].astype(pd.Int32Dtype())
#        if 'outfilename' not in self.imoutcols: self.imoutcols.append('outfilename')
#        if 'dl_code' not in self.imoutcols: self.imoutcols.append('dl_code')
#        if 'dl_str' not in self.imoutcols: self.imoutcols.append('dl_str')
        for ix in ix_selected_products:
            self.mk_outfilename(productTable,
                                ix,
                                outdir=outdir,
                                skip_propID2outsubdir=skip_propID2outsubdir,
                                obsnum2outsubdir=obsnum2outsubdir,
                                jpg_separate_subdir=jpg_separate_subdir,
                                skip_check_if_outfile_exists=skip_check_if_outfile_exists,
                                skip_check_filesize=skip_check_filesize)
        return(self.ix_selected_products)


    def update_obsTable_with_selectedProducts(self, obsTable=None, productTable=None, ix_selected_products=None, filetypes=None):
        """
        links product and observation Table. Counts how many products of which type for each observation.

        Parameters
        ----------
        obsTable : pdastroclass, optional
            Table with observations. The default is None. If None, then self.obsTable is used
        productTable : pdastroclass, optional
            Table with products. The default is None. If None, then self.productTable is used
        ix_selected_products : indeces, optional
            indices of selected products for productTable. The default is None. If None, then self.ix_selected_products is used

        Returns
        -------
        None.

        """
        if obsTable is None:
            obsTable=self.obsTable

        if productTable is None:
           productTable=self.productTable

        if ix_selected_products is None:
            ix_selected_products = self.ix_selected_products

        if filetypes is None:
            filetypes=self.params['filetypes']
        # add the filetypes to the output columns of obsTable
        for filetype in filetypes:
            if filetype not in self.params['outcolumns_obsTable']:
                self.params['outcolumns_obsTable'].append(filetype)

        for filetype in filetypes:
            obsTable.t[filetype]=None

        ixs_obsTable = obsTable.getindices()
        for ix_obsTable in ixs_obsTable:
            # find all products for a given observation
            obsid = obsTable.t.loc[ix_obsTable,'obsid']
            if self.verbose>2:
                print('### obsid',obsid)
            ixs_prodTable = productTable.ix_equal('parent_obsid',obsid,indices=ix_selected_products)
            if self.verbose>2:
                productTable.write(indices=ixs_prodTable)

            # count the product filetypes for each observations
            for ix_prodTable in ixs_prodTable:
                if obsTable.t.loc[ix_obsTable,productTable.t.loc[ix_prodTable,'filetype']] is None:
                    obsTable.t.loc[ix_obsTable,productTable.t.loc[ix_prodTable,'filetype']] = 0
                obsTable.t.loc[ix_obsTable,productTable.t.loc[ix_prodTable,'filetype']] += 1

        return(0)

    def mk_summary_tables(self, obsTable=None, ix_obs_sorted=None, filetypes=None):
        if obsTable is None:
            obsTable=self.obsTable

        if ix_obs_sorted is None:
            ix_obs_sorted = self.ix_obs_sorted


        if filetypes is None:
            filetypes=self.params['filetypes']

        # Add the filetypes to the summary table column
        columns_summary = ['proposal_id','obsnum']
        columns_summary.extend(filetypes)
        columns_summary.extend(['date_start'])

        self.summary = pdastroclass(columns=columns_summary)
        self.summary.t['obsnum']=pd.Series(dtype=pd.Int32Dtype())
        self.summary.t['proposal_id']=pd.Series(dtype=pd.Int64Dtype())

        #Loop through each propID
        propIDs = unique(obsTable.t.loc[ix_obs_sorted,'proposal_id'])
        for propID in propIDs:
            ixs_propID =  obsTable.ix_equal('proposal_id',propID,indices=ix_obs_sorted)

            ixs_propID_notnull = obsTable.ix_not_null('obsnum', indices=ixs_propID)
            obsnums = unique(obsTable.t.loc[ixs_propID_notnull,'obsnum'])
            if len(ixs_propID)!=len(ixs_propID_notnull):
                obsnums.append(pd.NA)
            if self.verbose>2:
                print('\n#### propID:',propID,' obsnums:',obsnums)
            # Loop through echo obsnum for the given propID, and make a summary entry
            for obsnum in obsnums:
                newline_obsnum={'proposal_id':propID}
                if obsnum is pd.NA:
                    ixs_obsnum = obsTable.ix_is_null('obsnum',indices=ixs_propID)
                else:
                    ixs_obsnum = obsTable.ix_equal('obsnum',obsnum,indices=ixs_propID_notnull)
                if self.verbose>2:
                    obsTable.write(indices=ixs_obsnum)
                newline_obsnum['obsnum']=obsnum

                #if obsnum is None:
                #    newline_obsnum['obsnum']=None
                #else:
                #    newline_obsnum['obsnum']=obsnum

                for filetype in filetypes:
                    newline_obsnum[filetype]=obsTable.t.loc[ixs_obsnum,filetype].sum()

                # get the start and end date for an observation
                date_start = obsTable.t.loc[ixs_obsnum,'date_min'].min()
                newline_obsnum['date_start']=date_start

                self.summary.newrow(newline_obsnum)

        #make the type integer for the filetype, proposal_id, and obsnum columns
        for filetype in filetypes:
            self.summary.t[filetype] = self.summary.t[filetype].astype('int')
        self.summary.t['proposal_id'] = self.summary.t['proposal_id'].astype('int')


        #self.summary.t['obsnum'] = self.summary.t['obsnum'].astype('int')

        # sort summary table
        self.ix_summary_sorted = self.summary.ix_sort_by_cols(self.params['sortcols_summaryTable'])

        return(0)

    def TNjpgname(self,jpgname):
        TNname = re.sub('jpg$','TN.jpg',jpgname)
        if TNname==jpgname:
            raise RuntimeError(f'thumbnail name {TNname} is equal to the original name {jpgname}!')
        return(TNname)

    def mk_thumbnail(self, jpgname, TNjpgname, tn_width, tn_height):
        #basewidth = 300
        img = Image.open(jpgname)
        if tn_width is None and tn_height is None:
            raise RuntimeError('Cannot make thumbnail, at least one of width and height has to be specified')
        if tn_height is None:
            wpercent = (tn_width / float(img.size[0]))
            tn_height = int((float(img.size[1]) * float(wpercent)))
        elif tn_width is None:
            hpercent = (tn_height / float(img.size[1]))
            tn_width = int((float(img.size[0]) * float(hpercent)))

        img = img.resize((tn_width, tn_height), Image.ANTIALIAS)
        img.save(TNjpgname)


    def mk_thumbnails(self, productTable, ixs,
                      jpg_separate_subdir=None,figcols=None,
                      tn_width=None, tn_height=None, overwrite=None):

        if tn_width is None:
            tn_width=self.params['webpage_thumbnails_width']

        if tn_height is None:
            tn_height=self.params['webpage_thumbnails_height']

        if jpg_separate_subdir is None:
            jpg_separate_subdir=self.params['jpg_separate_subdir']

        if figcols is None:
           figcols =self.params['webpage_level12_jpgs']

        if overwrite is None:
           overwrite =self.params['webpage_thumbnails_overwrite']

        ixs_uncal  = productTable.ix_equal('filetype','_uncal.fits',indices=ixs)

        if len(ixs_uncal)==0:
            print('WARNING: no entries to do jpg thumbnails for! returning ')
            return(0)

        suffixes = []
        for figcol in figcols:
            suffix = figcol
            if not re.search('^_',suffix): suffix=f'_{suffix}'
            if not re.search('\.jpg$',suffix): suffix+='.jpg'
            suffixes.append(suffix)

        #make the thumbnails
        print(f'Making thumbnails for {suffixes}')
        for ix in ixs_uncal:
            outdir = os.path.dirname(productTable.t.loc[ix,'outfilename'])
            if jpg_separate_subdir: outdir += '/jpg'
            for suffix in suffixes:
                jpgname = f"{outdir}/{productTable.t.loc[ix,'obs_id']}{suffix}"
                TNjpgname = self.TNjpgname(jpgname)
                if overwrite or not os.path.isfile(TNjpgname):
                    if os.path.isfile(jpgname):
                        if self.verbose>2: print(f'Making thumbnail {TNjpgname} for {jpgname}')
                        self.mk_thumbnail(jpgname, TNjpgname, tn_width, tn_height)
        print('Making thumbnails done!')
        return(0)

    def mk_webpage4ixs(self, ixs, productTable,
                       htmlname,asciiname,description,
                       jpg_separate_subdir=None,
                       width=None, height=None,
                       tn_width=None, tn_height=None,
                       figcols=None, fitskeys2table=None,
                       sortcols=None, mkthumbnails=None):

        if width is None:
            width=self.params['webpage_tablefigsize_width']

        if height is None:
            height=self.params['webpage_tablefigsize_height']

        if jpg_separate_subdir is None:
            jpg_separate_subdir=self.params['jpg_separate_subdir']

        if figcols is None:
           figcols =self.params['webpage_level12_jpgs']

        if fitskeys2table is None:
           fitskeys2table = self.params['webpage_fitskeys2table']

        if sortcols is None:
           sortcols =self.params['webpage_sortcols']

        if mkthumbnails is None:
           mkthumbnails =self.params['webpage_mkthumbnails']

        ixs_uncal  = productTable.ix_equal('filetype','_uncal.fits',indices=ixs)
        ixs_uncal =  productTable.ix_sort_by_cols(sortcols,indices=ixs_uncal)

        if len(ixs_uncal)==0:
            print(f'WARNING: no entries for {description}! Removing {htmlname}, {asciiname}, and returning...')
            rmfile(htmlname)
            rmfile(asciiname)
            return(0)

        suffixes = []
        for figcol in figcols:
            suffix = figcol
            if not re.search('^_',suffix): suffix=f'_{suffix}'
            if not re.search('\.jpg$',suffix): suffix+='.jpg'
            suffixes.append(suffix)


        # copy the fitsheader keywords to the table
        productTable.fitsheader2table('outfilename',indices=ixs_uncal,
                                      optionalfitskey=fitskeys2table)

        if mkthumbnails:
            self.mk_thumbnails(productTable, ixs_uncal,
                               jpg_separate_subdir=jpg_separate_subdir,figcols=figcols,
                               tn_width=tn_width, tn_height=tn_height)

        #add the figures to the table
        for ix in ixs_uncal:
            outdir = os.path.dirname(productTable.t.loc[ix,'outfilename'])
            if jpg_separate_subdir: outdir += '/jpg'

            # get the subdir of outdir that is not common with the html dir
            (commonpath,subdir)=split_commonpath(outdir, os.path.dirname(htmlname))
            if commonpath == '': raise RuntimeError(f'Somethings is wrong, htmldir={os.path.dirname(htmlname)} has no common path with outdir={outdir}')

            for suffix,figcol in zip(suffixes,figcols):
                if  subdir=='':
                    jpgname = f"{productTable.t.loc[ix,'obs_id']}{suffix}"
                else:
                    jpgname = f"{subdir}/{productTable.t.loc[ix,'obs_id']}{suffix}"

                if mkthumbnails:
                    TNjpgname = self.TNjpgname(jpgname)
                else:
                    TNjpgname=None

                # make a thumbnail that links to the full size image
                productTable.t.loc[ix,figcol]=image4table(jpgname,TNimagename=TNjpgname,width=width,height=height)
                #productTable.t.loc[ix,figcol]=addlink2string(imagestring4web(jpgname,width=None,height=p),jpgname)

        outcols4html = self.params['webpage_cols4table']

        # Make the ascci table(without the jpg cols), and save it.
        # Use a try/except to insulate against the case where requested suffixes for download in webpage_level12_jpgs
        # are not included in webpage_cols4table
        outcols4ascci = copy.deepcopy(outcols4html)
        for figcol in figcols:
            try:
                outcols4ascci.remove(figcol)
            except ValueError:
                pass
        print(f'Writing {description} pandas ascii to {asciiname}')
        print(outcols4ascci)
        productTable.write(filename=asciiname, indices=ixs_uncal, columns=outcols4ascci)

        # write the table to index.html
        print(f'writing {description} html to {htmlname}')
        f=open(htmlname,'w')
        s_asciilink = addlink2string('ascii-table',os.path.basename(asciiname))
        f.writelines([f'Level 1+2 Products for {description} ({s_asciilink} here)'])
        (errorflag,lines)=productTable.write(return_lines=True, indices=ixs_uncal, columns=outcols4html, htmlflag=True, htmlsortedtable=True, escape=False)
        if errorflag: raise RuntimeError(f'Soemthing went wrong when doing the webpage table for {description}')
        f.writelines(lines)
        f.close()

        # Make sure sortable.js is in the html directory
        htmldir = os.path.dirname(htmlname)
        jsfilename = os.path.join(os.path.realpath(os.path.dirname(os.path.join('../', __file__))), 'sortable.js')
        dest_jsfilename = f'{htmldir}/sortable.js'
        if not os.path.isfile(dest_jsfilename):
            if not os.path.isfile(jsfilename):
                raise RuntimeError(f'java script {jsfilename} for sortable tables does not exist!')
            shutil.copy(jsfilename, dest_jsfilename)

        return(0)

    def mk_webpages4propIDs(self, productTable=None, ix_selected_products=None,
                            jpg_separate_subdir=None,
                            width=None, height=None,
                            figcols=None, fitskeys2table=None,
                            sortcols=None, mkthumbnails=None):
        if productTable is None:
            # make a deep copy so that the thumbnail columns are not copied into the self.productTable
            productTable= copy.deepcopy(self.productTable)

        if ix_selected_products is None:
            ix_selected_products = self.ix_selected_products

        propIDs = unique(productTable.t.loc[ix_selected_products,'proposal_id'])
        for propID in propIDs:
            if self.verbose:
                print(f'Making webpage for propID={propID}')
            propID = int(propID)
            htmldir = f'{self.outrootdir}/{propID:05d}'
            htmlname = f'{htmldir}/index.html'
            asciiname = f'{htmldir}/{propID:05d}.products.txt'
            # description is used for the title of the webpage etc
            description=f'propID={propID}'
            # get all indices for propID
            ixs_propID = productTable.ix_equal('proposal_id',propID,indices=ix_selected_products)

            self.mk_webpage4ixs(ixs_propID, productTable,
                                htmlname,asciiname,description,
                                jpg_separate_subdir=jpg_separate_subdir,
                                width=width, height=height,
                                figcols=figcols, fitskeys2table=fitskeys2table,
                                sortcols=sortcols,mkthumbnails=mkthumbnails)


    def mk_all_tables(self, filetypes=None, showtables=True):

        if filetypes is not None:
            self.params['filetypes'] = filetypes

        # get the MJD limits, based on --mjdlimits --lockbacktime --mjdmin --mjdmax
        mjd_min, mjd_max = self.get_mjd_limits()

        # get the observations: stored in self.obsTable
        self.observation_query(propID=self.params['propID'], mjd_min = mjd_min, mjd_max = mjd_max)

        if len(self.obsTable.t)==0:
            print('\n################################\nNO OBSERVATIONS FOUND! exiting....\n################################')
            self.ix_selected_products = []
            return(0)

        if self.verbose>1:
            print(self.obsTable.t)

        # get the products:  stored in self.productTable
        # this list contains in general several entries for each observations.
        # If not otherwise specified, uses self.obsTable as starting point
        self.product_query(Nobs_per_batch=self.params['Nobs_per_batch'])
        if self.verbose>1:
            print(self.productTable.t)

        # select the products to download
        # If not otherwise specified, uses self.productTable as starting point
        # uses self.filetypes for filtering, which is set by optional parameters.
        # the selected products indices are put into self.ix_selected_products
        self.product_filter()

        # get selected obsnums if specified
        self.ix_selected_products = self.select_obsnums(self.params['obsnums'])

        # make some obsid cuts!
        self.ix_selected_products = self.obsid_select(self.params['obsid_select'])
        self.ix_selected_products = self.obsid_list(self.params['obsid_list'])

        # get selected SCAs if specified
        self.ix_selected_products = self.select_sca(self.params['sca'])

        # only keep entries in the obstable that are parent_obsid in product table
        self.ix_obs_sorted = self.update_obstable_indices(self.ix_selected_products)

        # define the output filenames, and check if they exist.
        self.mk_outfilenames(skip_propID2outsubdir=self.params['skip_propID2outsubdir'],
                             obsnum2outsubdir=self.params['obsnum2outsubdir'],
                             jpg_separate_subdir=self.params['jpg_separate_subdir'],
                             skip_check_if_outfile_exists=self.params['skip_check_if_outfile_exists'],
                             skip_check_filesize=self.params['skip_check_filesize'])

        # update obsTable with selected products: count the different filetypes, and also update the obsnum if None in obsTable
        self.update_obsTable_with_selectedProducts()

        # print the table to screen if either verbose or not downloading
        if showtables:
            print('\n######################\n### Selected Products:\n######################')
            self.productTable.write(indices=self.ix_selected_products, columns=self.params['outcolumns_productTable'])

            print('\n#################\n### Observations:\n#################')
            self.obsTable.write(columns=self.params['outcolumns_obsTable'],indices=self.ix_obs_sorted)

        # make summary table
        self.mk_summary_tables()
        if showtables:
            print('\n##########################\n### Summary propID/obsnum:\n##########################')
            self.summary.write(indices=self.ix_summary_sorted)

        # save the tables if wanted
        if self.params['savetables'] is not None:
            prodfile = os.path.join(self.outrootdir, self.params['savetables']+'.selprod.txt')
            obsfile = os.path.join(self.outrootdir, self.params['savetables']+'.obs.txt')
            summaryfile = os.path.join(self.outrootdir, self.params['savetables']+'.summary.txt')

            self.productTable.write(filename=prodfile, indices=self.ix_selected_products, columns=self.params['outcolumns_productTable'], verbose=2)
            self.obsTable.write(filename=obsfile, columns=self.params['outcolumns_obsTable'], indices=self.ix_obs_sorted, verbose=2)
            self.summary.write(filename=summaryfile, indices=self.ix_summary_sorted, verbose=2)

        return(0)



if __name__ == '__main__':

    query = query_mast()
    parser = query.define_options()
    args = parser.parse_args()

    # the arguments are saved in query.params
    query.get_arguments(args)
    if query.verbose>2:
        print('params:', query.params)

    # use arguments or $API_MAST_TOKEN to login
    query.login(raiseErrorFlag=False)

    # self.outrootdir is set depending on outrootdir and outsubdir in cfg file or through the options --outrootdir and --outsubdir
    query.set_outrootdir()
    if query.verbose: print(f'Outdir: {query.outdir}')

    # make the tables
    query.mk_all_tables()

    # make the webpages
    if args.makewebpages:
        query.mk_webpages4propIDs()






#!/usr/bin/env python
import os
import pkgutil
import subprocess
import sys
from setuptools import setup, find_packages, Extension, Command
from setuptools.command.test import test as TestCommand


# allows you to build sphinx docs from the package
# main directory with "python setup.py build_sphinx"

try:
    from sphinx.cmd.build import build_main
    from sphinx.setup_command import BuildDoc

    class BuildSphinx(BuildDoc):
        """Build Sphinx documentation after compiling C source files"""

        description = 'Build Sphinx documentation'

        def initialize_options(self):
            BuildDoc.initialize_options(self)

        def finalize_options(self):
            BuildDoc.finalize_options(self)

        def run(self):
            build_cmd = self.reinitialize_command('build_ext')
            build_cmd.inplace = 1
            self.run_command('build_ext')
            build_main(['-b', 'html', './docs', './docs/_build/html'])

except ImportError:
    class BuildSphinx(Command):
        user_options = []

        def initialize_options(self):
            pass

        def finalize_options(self):
            pass

        def run(self):
            print('!\n! Sphinx is not installed!\n!', file=sys.stderr)
            exit(1)

DOCS_REQUIRE = [
    'nbsphinx',
    'sphinx',
    'sphinx-automodapi',
    'sphinx-rtd-theme',
    'stsci-rtd-theme',
    'extension-helpers',
]
TESTS_REQUIRE = [
    'pytest',
]

setup(
    name='jwst_mast_query',
    description='Query MAST for JWST products, and download them',
    long_description=('Query MAST for JWST observations, select certain types of these products, provide a summary of these observations and products, and optionally download the selected products'),
    author='STScI (Rest)',
    author_email='arest@stsci.edu',
    url='https://github.com/spacetelescope/jwst_mast_query',
    keywords=['astronomy'],
    classifiers=[
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: BSD License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Topic :: Scientific/Engineering :: Astronomy',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ],
    packages = find_packages(),
    package_data={},
    use_scm_version=True,
    setup_requires=['setuptools_scm'],
    install_requires=[
        'astroquery>=0.4.2',
        'astropy',
        'ipython',
        'jinja2',
        'numpy',
        'pandas',
        'pyyaml',
        'scipy',
        'argparse',
        'pillow'
    ],
    scripts=['jwst_mast_query/scripts/jwst_download.py'],
    include_package_data=True,
    extras_require={
        'docs': DOCS_REQUIRE,
        'test': TESTS_REQUIRE,
    },
    tests_require=TESTS_REQUIRE,
    cmdclass={
        'build_sphinx': BuildSphinx
    },)



"""Globally defined and used variables for the ``jwst_mast_query`` project.

"""
import os


MAST_OBS_MODES = {'miri': ['MIRI/CORON', 'MIRI/IFU', 'MIRI/IMAGE', 'MIRI/SLIT', 'MIRI/SLITLESS', 'MIRI/TARGACQ'],
                  'nircam': ['NIRCAM/CORON', 'NIRCAM/GRISM', 'NIRCAM/IMAGE', 'NIRCAM/TARGACQ', 'NIRCAM/WFSS'],
                  'niriss': ['NIRISS/AMI', 'NIRISS/IMAGE', 'NIRISS/SOSS', 'NIRISS/WFSS'],
                  'nirspec': ['NIRSPEC/BOTS', 'NIRSPEC/IFU', 'NIRSPEC/MSA', 'NIRSPEC/SLIT']
                  }

# Default for config file, if available
if 'JWST_QUERY_CFGFILE' in os.environ and os.environ['JWST_QUERY_CFGFILE'] != '':
    cfgfilename = os.environ['JWST_QUERY_CFGFILE']
else:
    cfgfilename = None

# Default for token is $MAST_API_TOKEN
if 'MAST_API_TOKEN' in os.environ:
    defaulttoken = os.environ['MAST_API_TOKEN']
else:
    defaulttoken = None

PARAM_DEFAULTS = {'mastcolumns_obsTable': ['proposal_id','dataURL','obsid','obs_id','t_min','t_exptime','instrument_name'],
                  'outcolumns_productTable': ['proposal_id','obsnum','obsID','parent_obsid','obs_id','sca','visit','dataproduct_type','productFilename','filetype','calib_level','size','description'],
                  'outcolumns_obsTable': ['proposal_id','obsnum','obsid','obs_id','t_min','t_exptime','date_min','instrument_name'],
                  'sortcols_productTable': ['calib_level','filetype','obsID'],
                  'sortcols_obsTable': ['date_min','proposal_id','obsnum'],
                  'sortcols_summaryTable': ['date_start','proposal_id','obsnum'],
                  'configfile': cfgfilename,
                  'token': defaulttoken,
                  'instrument': None,
                  'obsmode': None,
                  'propID': None,
                  'obsnums': None,
                  'obsid_select': [],
                  'obsid_list': [],
                  'sca': None,
                  'verbose': 0,
                  'filetypes': ['fits'],
                  'guidestars': False,
                  'lookbacktime': 1.,
                  'calib_levels': None,
                  'Nobs_per_batch': 2,
                  'obsnum2outsubdir': None,
                  'propIDs_obsnum2outsubdir': [],
                  'date_select': [],
                  'savetables': None,
                  'makewebpages': False,
                  'outrootdir': None,
                  'outsubdir': None,
                  'clobber': False,
                  'skip_propID2outsubdir': False,
                  'skip_check_if_outfile_exists': False,
                  'skip_check_filesize': False,
                  'jpg_separate_subdir': False,
                  'webpage_tablefigsize_width': None,
                  'webpage_tablefigsize_height': None,
                  'webpage_level12_jpgs': ['_uncal.jpg', '_dark.jpg', '_rate.jpg', '_rateints.jpg', '_trapsfilled.jpg', '_cal.jpg', '_crf.jpg'],
                  'webpage_fitskeys2table': ['TARG_RA', 'TARG_DEC', 'FILTER', 'PUPIL', 'READPATT', 'NINTS', 'NGROUPS', 'NFRAMES', 'DATE-BEG',
                                             'DATE-END', 'EFFINTTM', 'EFFEXPTM'],
                  'webpage_cols4table': ['proposal_id', 'obsnum', 'visit', 'obsID', 'parent_obsid', 'sca', 'FILTER', 'PUPIL', 'READPATT', 'uncal',
                                         'dark', 'rate', 'rateints', 'cal', 'TARG_RA', 'TARG_DEC', 'NINTS', 'NGROUPS', 'NFRAMES', 'DATE-BEG', 'DATE-END',
                                         'EFFINTTM', 'EFFEXPTM', 'size', 'obs_id', 'outfilename'],
                  'webpage_sortcols': ['proposal_id', 'obsnum', 'visit', 'sca'],
                  'webpage_mkthumbnails': True,
                  'webpage_thumbnails_overwrite': False,
                  'webpage_thumbnails_width': 120,
                  'webpage_thumbnails_height': None,
                  'login': None,
                  'skipdownload': False
                  }